from .parser_kodik import KodikParser
from .parser_kodik_async import KodikParserAsync
from .parser_aniboom import AniboomParser
from .parser_aniboom_async import AniboomParserAsync
from .parser_jutsu import JutsuParser
from .parser_shikimori import ShikimoriParser
from .parser_shikimori_async import ShikimoriParserAsync
from .api_kodik import KodikList, KodikSearch


import requests
try:
    import lxml
except ImportError:
    LXML_WORKS = False
else:
    LXML_WORKS = True
from bs4 import BeautifulSoup as Soup

try:
    from . import errors # Импорт если библиотека установлена
except ImportError:
    import errors # Импорт если библиотека не установлена и файл лежит локально

class JutsuParser:
    """
    Парсер jut.su
    """
    def __init__(self, use_lxml: bool = False, mirror: str|None = None) -> None:
        """
        :use_lxml: Использовать lxml парсер. В некоторых случаях может не работать, однако работает значительно быстрее стандартного.
        :mirror: В случае, если оригинальный домен заблокирован, можно использовать этот параметр, чтобы заменить адрес сайта на зеркало. Пример: "1234.net"
        """
        if not LXML_WORKS and use_lxml:
            raise ImportWarning('Параметр use_lxml установлен в true, однако при попытке импорта lxml произошла ошибка')
        self.USE_LXML = use_lxml
        if mirror: # Если есть зеркало, то меняем домен на него
            self._dmn = mirror
        else:
            self._dmn = "jut.su"
    
    def get_mp4_link(self, link: str) -> dict:
        """
        Получает ссылку на mp4 файл по ссылке на страницу.

        :link: ссылка на страницу c плеером (прим: https://jut.su/tondemo-skill/episode-1.html)
        
        Возвращает словарь в виде:
        {
            'качество': 'ссылка'
        }
        """
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:127.0) Gecko/20100101 Firefox/127.0',
        }
        response = requests.get(link, headers=headers)
        if response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.text
        soup = Soup(response, 'lxml') if self.USE_LXML else Soup(response, 'html.parser')
        if soup.find('video', {'id': 'my-player'}) is None:
            raise errors.UnexpectedBehavior(f'Ожидался тег "video" на странице по ссылке "{link}"')
        sources = soup.find('video', {'id': 'my-player'}).find_all('source')
        links = {}
        for src in sources:
            q = src.get_attribute_list('res')[0]
            link = src.get_attribute_list('src')[0]
            links[q] = link
        return links
    
    def get_anime_info(self, link: str) -> dict:
        """
        Получает данные аниме по ссылке.

        :link: ссылка на страницу аниме (прим: https://jut.su/tondemo-skill/)

        Возвращает словарь вида:
        {
            "title": "Название аниме",
            "origin_title": "Оригинальное название (транслит японского названия на английском)",
            "age_rating": "Возрастное ограничение",
            "description": "Описание",
            "years": ["Год выхода 1 сезона", "Год выхода 2 сезона"],
            "genres": ["Жанр 1", "Жанр 2"],
            "poster": "Ссылка на картинку (плохое качество)",
            "seasons": [
                [ # 1 сезон будет обязательно, даже если у аниме нет других сезонов
                    "ссылка на 1 серию 1 сезона (страница с плеером)",
                    "ссылка на 2 серию 1 сезона (страница с плеером)"
                ],
                [ # 2 сезон если есть
                    "ссылка на 1 серию 2 сезона (страница с плеером)",
                    "ссылка на 2 серию 2 сезона (страница с плеером)"
                ],
                ...
            ],
            "seasons_names": [ # Если у аниме только 1 сезон, этот список будет пустым
                "Название 1 сезона", 
                "Название 2 сезона"
            ],
            "films": [ # Если фильмов нет - список пустой
                "Ссылка на фильм 1 (страница с плеером)",
                "Ссылка на фильм 2 (страница с плеером)",
            ]
        }
        """
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:127.0) Gecko/20100101 Firefox/127.0',
        }
        response = requests.get(link, headers=headers)
        if response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.text
        soup = Soup(response, 'lxml') if self.USE_LXML else Soup(response, 'html.parser')
        soup = soup.find('div', {'id': 'dle-content'})

        if soup.find('a', {'class': 'video'}) is None:
            raise errors.UnexpectedBehavior(f'Предполагалось, что на странице "{link}" будет находится информация об аниме и ссылки на серии, однако они были не найдены. Проверьте ссылку вручную.')

        c_data = {}
        c_data['title'] = soup.find('h1', {'class': 'header_video'}).text.replace('Смотреть ', '', 1).replace(' все серии и сезоны', '', 1).replace(' все серии', '', 1)
        
        poster = soup.find('div', {'class': 'all_anime_title'})
        if poster is None:
            c_data['poster'] = None
        else:
            poster = poster.get_attribute_list('style')[0]
            c_data['poster'] = poster[poster.find("'")+1:poster.rfind("'")]

        additional_data = soup.find('div', {'class': 'under_video_additional'})
        c_data['genres'] = []
        c_data['years'] = [] # Потому что может быть указано сразу несколько годов (для разных сезонов)
        for a in additional_data.find_all('a'):
            href = a.get_attribute_list('href')[0]
            content = a.get_text()
            content = content[content.find(' ')+1:]
            if href[href.rfind('/', 0, -2)+1:-1].isdigit() or \
                all(x.isdigit() for x in href[href.rfind('/', 0, -2)+1:-1].split('-')):
                c_data['years'].append(content)
            else:
                c_data['genres'].append(content)

        c_data['original_title'] = additional_data.find('b').text
        c_data['age_rating'] = additional_data.find('span', {'class': 'age_rating_all'}).text
        
        desc = soup.find('p')
        for i in desc.find_all('i'):
            i.decompose() # Чистим текст от всякого мусора
        c_data['description'] = desc.text.strip()

        c_data['seasons'] = []
        c_data['films'] = []
        # А тут начинается веселье с эпизодами и сезонами, потому что на jutsu оно в коде никак не выделяется отдельно. Поэтому следите за руками
        for episode in soup.find_all('a', {'class': 'video'}):
            href = episode.get_attribute_list('href')[0]
            if 'season' in href:
                season_num = int(href[href.find('season')+7:href.rfind('/')])
                if len(c_data['seasons']) < season_num:
                    c_data['seasons'].append([])
                c_data['seasons'][season_num-1].append(f'https://{self._dmn}/'+href)
            elif 'film' in href:
                c_data['films'].append(f'https://{self._dmn}/'+href)
            else:
                if len(c_data['seasons']) == 0:
                    c_data['seasons'].append([])
                c_data['seasons'][-1].append(f'https://{self._dmn}/'+href)

        # Если сезонов как минимум 2, то скорее всего у них есть название (либо ремейки, они тоже как отдельный сезон идут)
        c_data['seasons_names'] = []
        if len(c_data['seasons']) > 1:
            for season_name in soup.find_all('h2', {'class': 'the-anime-season'}):
                c_data['seasons_names'].append(season_name.text.strip())

        return c_data
    
    """
    Поиска нет, потому что jut.su использует яндекс поиск для поиска по своему сайту.
    И это поиск настолько непредсказуем, что проще не трогать.
    В качестве поиска вы можете использовать оригинальное название для ссылки.
    Типо так:
    
    Название аниме: Волчица и пряности
    Оригинальное название: Ookami to Koushinryou
    Ссылка на страницу: https://jut.su/ookami-to-koshinryou/

    Ну суть понятна.
    Правда есть и исключения по типу наруто, которое в принципе не подчиняется правилам других аниме.
    """


try:
    import aiohttp
    import asyncio
except ImportError:
    CAN_WORK = False
else:
    CAN_WORK = True

import json

class Response:
    """
    Класс для удобства использования AsyncSession
    Параметры:
    status, status_code, text, url
    """
    def __init__(self, status, text, url) -> None:
        self.status = status
        self.status_code = status
        self.text = text
        self.url = url

    def json(self):
        return json.loads(self.text)

class AsyncSession:
    """
    Класс-обертка для удобства использования асинхронных запросов
    """
    def __init__(self) -> None:
        if not CAN_WORK:
            raise ImportError("Невозможно получить доступ к библиотекам \"aiohttp\" и/или \"asyncio\". Проверьте правильность установки библиотеки.")

    async def get(self, url: str, **kwargs) -> Response:
        async with aiohttp.request(method='get', url=url, **kwargs) as response:
            text = await response.text()
            res = Response(
                status=response.status,
                text=text,
                url=response.url.human_repr()
            ) 
        return res
    
    async def post(self, url: str, **kwargs) -> Response:
        async with aiohttp.request(method='post', url=url, **kwargs) as response:
            text = await response.text()
            res = Response(
                status=response.status,
                text=text,
                url=response.url.human_repr()
            )
        return res


class TokenError(Exception):
    """
    Ошибка для обозначения неверного токена.
    """

class ServiceError(Exception):
    """
    Ошибка для обозначения ошибки на стороне сервера
    """

class PostArgumentsError(Exception):
    """
    Ошибка для обозначения неверно переданных аргументов серверу
    """

class NoResults(Exception):
    """
    Ошибка для обозначения отсутствия результатов
    """

class UnexpectedBehavior(Exception):
    """
    Ошибка для обозначения неожиданного или необработанного поведения
    """

class QualityNotFound(Exception):
    """
    Ошибка для обозначения не найденного запрашиваемого качества видео
    """

class AgeRestricted(Exception):
    """
    Ошибка для обозначения что контент заблокирован из-за возрастного рейтинга
    """

class TooManyRequests(Exception):
    """
    Ошибка для обозначения ошибки 429 из-за слишком частых запросов.
    В основном для шикимори
    """

class ContentBlocked(Exception):
    """
    Ошибка для обозначения заблокированного контента/плеера
    """

class ServiceIsOverloaded(Exception):
    """
    Ошибка для обозначения http кода 520
    Используется в парсере shikimori
    """


try:
    import lxml
except ImportError:
    LXML_WORKS = False
else:
    LXML_WORKS = True
import re
from bs4 import BeautifulSoup as Soup
import json

try:
    from . import errors # Импорт если библиотека установлена
    from .internal_tools import AsyncSession
except ImportError:
    import errors # Импорт если библиотека не установлена и файл лежит локально
    from internal_tools import AsyncSession

class ShikimoriParserAsync:
    """
    Парсер шикимори. Не использует встроенный в шикимори api.
    """

    genres_list = ['1-Action', '2-Adventure', '3-Racing', '4-Comedy', '5-Avant-Garde', '6-Mythology', '7-Mystery', '8-Drama', '9-Ecchi', '10-Fantasy', '11-Strategy-Game', '13-Historical', '14-Horror', '15-Kids', '17-Martial-Arts', '18-Mecha', '19-Music', '20-Parody', '21-Samurai', '22-Romance', '23-School', '24-Sci-Fi', '25-Shoujo', '27-Shounen', '29-Space', '30-Sports', '31-Super-Power', '32-Vampire', '35-Harem', '36-Slice-of-Life', '37-Supernatural', '38-Military', '39-Detective', '40-Psychological', '42-Seinen', '43-Josei', '102-Team-Sports', '103-Video-Game', '104-Adult-Cast', '105-Gore', '106-Reincarnation', '107-Love-Polygon', '108-Visual-Arts', '111-Time-Travel', '112-Gag-Humor', '114-Award-Winning', '117-Suspense', '118-Combat-Sports', '119-CGDCT', '124-Mahou-Shoujo', '125-Reverse-Harem', '130-Isekai', '131-Delinquents', '134-Childcare', '135-Magical-Sex-Shift', '136-Showbiz', '137-Otaku-Culture', '138-Organized-Crime', '139-Workplace', '140-Iyashikei', '141-Survival', '142-Performing-Arts', '143-Anthropomorphic', '144-Crossdressing', '145-Idols-(Female)', '146-High-Stakes-Game', '147-Medical', '148-Pets', '149-Educational', '150-Idols-(Male)', '151-Romantic-Subtext', '543-Gourmet']

    def __init__(self, use_lxml: bool = False, mirror: str|None = None) -> None:
        """
        :use_lxml: Использовать lxml парсер. В некоторых случаях может не работать, однако работает значительно быстрее стандартного.
        :mirror: В случае, если оригинальный домен заблокирован, можно использовать этот параметр, чтобы заменить адрес сайта на зеркало. Пример: "1234.net"
        """
        self.requests = AsyncSession()

        if not LXML_WORKS and use_lxml:
            raise ImportWarning('Параметр use_lxml установлен в true, однако при попытке импорта lxml произошла ошибка')
        self.USE_LXML = use_lxml
        if mirror: # Если есть зеркало, то меняем домен на него
            self._dmn = mirror
        else:
            self._dmn = "shikimori.one"

    async def search(self, title: str) -> list:
        """
        Быстрый поиск аниме по названию (ограничено по количеству результатов).

        :title: название аниме

        Возвращает список словарей в виде:
        [
            {
                "genres": ["Жанр1", "Жанр2"],
                "link": "Ссылка на страницу аниме",
                "original_title": "Оригинальное название (транслит японского названия на английском)",
                "poster": "Ссылка на постер к аниме (плохое качество) (если есть, иначе None)",
                "shikimori_id": "id шикимори",
                "status": "статус (вышло, онгоинг, анонс) (если есть, иначе None)",
                "studio": "студия анимации (если есть, иначе None)",
                "title": "Название",
                "type": "тип аниме (TV сериал, OVA, ONA, ...) (если есть, иначе None)",
                "year": "год выхода (если есть, иначе None)"
            }
        ]
        """
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
            'Accept': 'application/json, text/plain, */*',
            'X-Requested-With': 'XMLHttpRequest',
        }

        params = {
            'search': title,
        }
        # Используем autocomplete эндпоинт, потому что обычный поиск тупо блокируется если находит 18+ контент
        response = await self.requests.get(f'https://{self._dmn}/animes/autocomplete/v2', params=params, headers=headers)
        if response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.json()['content']
        soup = Soup(response, 'lxml') if self.USE_LXML else Soup(response, 'html.parser')
        res = []
        for anime in soup.find_all('div', {'class': 'b-db_entry-variant-list_item'}):
            if anime.get_attribute_list('data-type')[0] != 'anime': # нас интересует только аниме
                continue
            c_data = {}
            c_data['link'] = anime.get_attribute_list('data-url')[0]
            c_data['shikimori_id'] = anime.get_attribute_list('data-id')[0]
            if not anime.find('div', {'class': 'image'}) is None:
                c_data['poster'] = anime.find('div', {'class': 'image'}).find('picture').find('img').get_attribute_list('srcset')[0].replace(' 2x', '')
            else:
                c_data['poster'] = None
            info = anime.find('div', {'class': 'info'})
            c_data['original_title'] = info.find('div', {'class': 'name'}).find('a').get_attribute_list('title')[0]
            c_data['title'] = info.find('div', {'class': 'name'}).find('a').text.split('/')[0]
            if info.find('div', {'class': 'line'}).find('div', {'class': 'key'}).text == 'Тип:':
                c_data['type'] = info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find('div', {'class': 'b-tag'}).text
                c_data['status'] = info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find_all('div', {'class': 'b-anime_status_tag'})[-1].get_attribute_list('data-text')[0]
                if len(info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find_all('div', {'class': 'b-anime_status_tag'})) > 1:
                    c_data['studio'] = info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find_all('div', {'class': 'b-anime_status_tag'})[0].get_attribute_list('data-text')[0]
                else:
                    c_data['studio'] = None
                if len(info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find_all('div', {'class': 'b-tag'})) > 1:
                    c_data['year'] = info.find('div', {'class': 'line'}).find('div', {'class': 'value'}).find_all('div', {'class': 'b-tag'})[-1].text.replace(' год', '')
                else:
                    c_data['year'] = None
            else:
                c_data['type'] = None # В некоторых случаях возможна ситуация когда не известна ни дата выхода, ни тип аниме.
                c_data['status'] = None
                c_data['studio'] = None
                c_data['year'] = None
            c_data['genres'] = []
            for genre in info.find_all('span', {'class': 'genre-ru'}):
                c_data['genres'].append(genre.text)
            res.append(c_data)
        return res
    
    async def anime_info(self, shikimori_link: str) -> dict:
        """
        Получение данных по аниме парсингом.

        :shikimori_link: ссылка на страницу шикимори с информацией (прим: https://shikimori.one/animes/z20-naruto)

        Возвращает словарь вида:
        {
            "dates": "Даты выхода",
            "description": "Описание",
            "episode_duration": "Средняя продолжительность серии",
            "episodes": "Количество эпизодов если статус 'вышло' или 'вышедших эпизодов / анонсировано эпизодов' или None (если фильм)",
            "genres": ["Жанр1", "Жанр2"],
            "licensed": "Кто лицензировал в РФ или None",
            "licensed_in_ru": "Название аниме как лицензировано в РФ или None",
            "next_episode": "Дата выхода следующего эпизода или None",
            "original_title": "Оригинальное название",
            "picture": "Ссылка на jpeg постер",
            "premiere_in_ru": "Дата премьеры в РФ или None",
            "rating": "возрастной рейтинг",
            "score": "оценка на шикимори",
            "status": "статус выхода",
            "studio": "студия анимации",
            "themes": ["Тема1", "Тема2"],
            "title": "Название на русском",
            "type": "тип аниме (TV Сериал, Фильм, т.п.)"
        }
        """
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
        }
        response = await self.requests.get(shikimori_link, headers=headers)
        if response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.text
        soup = Soup(response, 'lxml') if self.USE_LXML else Soup(response, 'html.parser')
        if not soup.find('p', {'class': 'age-restricted-warning'}) is None:
            raise errors.AgeRestricted(f'Аниме по ссылке "{shikimori_link}" невозможно обработать из-за блокировки по возрастному рейтингу.')
        res = {}
        title = soup.find('header', {'class': 'head'}).find('h1').text.split(' / ')
        res['title'] = title[0]
        res['original_title'] = title[1]
        picture = soup.find('picture')
        if not picture is None:
            res['picture'] = picture.find('img').get_attribute_list('srcset')[0].replace(' 2x', '')
        else:
            res['picture'] = None
        info = soup.find('div', {'class': 'c-info-left'}).find('div', {'class': 'block'})
        for line in info.find_all('div', {'class': 'line'}):
            key = line.find('div', {'class': 'key'}).text
            value = line.find('div', {'class': 'value'})
            if key == 'Тип:': res['type'] = value.text
            elif key == 'Эпизоды:': res['episodes'] = value.text
            elif key == 'Следующий эпизод:': res['next_episode'] = value.find('span').get_attribute_list('data-datetime')[0]
            elif key == 'Длительность эпизода:': res['episode_duration'] = value.text
            elif key == 'Статус:': 
                res['status'] = value.find('span').get_attribute_list('data-text')[0]
                if res['status'] == 'онгоинг' or res['type'] == 'Фильм':
                    res['dates'] = value.text.strip()
                elif len(value.find_all('span')) > 1:
                    res['dates'] = value.find_all('span')[-1].text
            elif key == 'Жанры:': res['genres'] = [genre.text for genre in value.find_all('span', 'genre-ru')]
            elif key == 'Темы:' or key == 'Тема:': res['themes'] = [genre.text for genre in value.find_all('span', 'genre-ru')]
            elif key == 'Рейтинг:': res['rating'] = value.text
            elif key == 'Лицензировано:': res['licensed'] = value.text
            elif key == 'Лицензировано в РФ под названием:': res['licensed_in_ru'] = value.text
            elif key == 'Премьера в РФ:': res['premiere_in_ru'] = value.text
        
        # Заполняем ключи если их нет в информации
        for k in ["type", "episodes", "next_episode", "episode_duration", "status", "genres", "themes", "rating", 
                  "dates", "licensed", "licensed_in_ru", "premiere_in_ru"]:
            if k not in res.keys():
                res[k] = None

        desc = soup.find('div', {'class': 'text'})
        if not desc is None:
            res['description'] = desc.text
        else:
            res['description'] = None
        score = soup.find('div', {'class': 'score-value'})
        if not score is None:
            res['score'] = score.text
        else:
            res['score'] = None
        
        studio = soup.find('a', {'title': re.compile(r'Аниме студии.*')})
        if not studio is None:
            res['studio'] = studio.get_attribute_list('title')[0].replace('Аниме студии ', '', 1)
        else:
            res['studio'] = None
        return res
    
    async def additional_anime_info(self, shikimori_link: str) -> dict:
        """
        Получение дополнительных данных об аниме.
        Получаемые данные: связанные аниме (продолжение, предыстория, альтернативное и т.п.), Авторы (автор манги, режиссер), Главные герои, Скриншоты, Ролики, Похожее
        
        :shikimori_link: ссылка на страницу шикимори с информацией (прим: https://shikimori.one/animes/z20-naruto)

        Возвращает словарь вида:
        {
            "related": [
                {
                    "date": "Даты выхода/сезон",
                    "name": "Название",
                    "picture": "Ссылка на картинку",
                    "relation": "тип связи (продолжение, предыстория, адаптация и т.п.)",
                    "type": "Тип (TV сериал, OVA, ONA, манга, ранобэ и т.д.)",
                    "url": "Ссылка на страницу шикимори"
                }
            ],
            "staff": [
                {
                    "name": "Имя человека (на русском)",
                    "roles": ["Роль1", "Роль2"],
                    "link": "ссылка шикимори на человека"
                }
            ],
            "main_characters": [
                {
                    "name": "Имя персонажа",
                    "picture": "Картинка (jpeg)"
                }
            ],
            "screenshots": ["Ссылка на скриншот 1", "Ссылка на скриншот 2"],
            "videos": [
                {
                    "name": "Название видео",
                    "link": "Ссылка на видео (обычно ютуб)"
                }
            ],
            "similar": [
                {
                    "name": "Название аниме (похожего)",
                    "picture": "Картинка (постер)",
                    "link": "Ссылка на шикимори"
                }
            ]
        }
        """

        link = shikimori_link+'resources' if shikimori_link[-1] == '/' else shikimori_link+'/resources'

        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
        }
        response = await self.requests.get(link, headers=headers)
        if response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.text
        soup = Soup(response, 'lxml') if self.USE_LXML else Soup(response, 'html.parser')
        if not soup.find('p', {'class': 'age-restricted-warning'}) is None:
            raise errors.AgeRestricted(f'Аниме по ссылке "{shikimori_link}" невозможно обработать из-за блокировки по возрастному рейтингу.')
        res = {
            'related': [], 'staff': [], 'main_characters': [], 'screenshots': [], 'videos': [], 'similar': []
        }
        r1 = soup.find('div', {'class': 'cc-related-authors'})
        for col in r1.find_all('div', {'class': 'c-column'}):
            col_type = col.find('div', {'class': 'subheadline'}).text
            if col_type == 'Связанное':
                for entry in col.find_all('div', {'class': 'b-db_entry-variant-list_item'}):
                    c_data = {}
                    c_data['url'] = entry.get_attribute_list('data-url')[0]
                    c_data['picture'] = None
                    if not entry.find('picture') is None:
                        c_data['picture'] = entry.find('picture').find('img').get_attribute_list('srcset')[0].replace(' 2x', '')
                    try:
                        c_data['name'] = entry.find('div', {'class': 'name'}).find('span', {'class': 'name-ru'}).text
                    except AttributeError:
                        try:
                            c_data['name'] = entry.find('div', {'class': 'name'}).find('span', {'class': 'name-en'}).text
                        except AttributeError:
                            c_data['name'] = None
                    for other in entry.find('div', {'class': 'line'}).find_all('div'):
                        cls = other.get_attribute_list('class')
                        if 'b-anime_status_tag' in cls:
                            c_data['relation'] = other.text
                        elif 'linkeable' in cls:
                            link = other.get_attribute_list('data-href')[0]
                            if '/kind/' in link:
                                c_data['type'] = other.text
                            elif '/season/' in link:
                                c_data['date'] = other.text
                    # Если в связанных клип (по другому отрисовано имя)
                    if c_data['type'] == 'Клип' and c_data['name'] == None:
                        try:
                            c_data['name'] = entry.find('div', {'class': 'name'}).find('a').text
                        except:
                            pass                
                    # Заполняем пустое
                    for k in ['relation', 'type', 'date']:
                        if k not in c_data.keys():
                            c_data[k] = None
                    res['related'].append(c_data)
            
            elif col_type == 'Авторы':
                for entry in col.find_all('div', {'class': 'b-db_entry-variant-list_item'}):
                    c_data = {}
                    c_data['link'] = entry.get_attribute_list('data-url')[0]
                    c_data['name'] = entry.get_attribute_list('data-text')[0]
                    c_data['roles'] = []
                    for role in entry.find('div', {'class': 'line'}).find_all('div', {'class': 'b-tag'}):
                        c_data['roles'].append(role.text)
                    res['staff'].append(c_data)

        r1 = soup.find('div', {'class': 'c-characters'})
        for char in r1.find_all('article'):
            c_data = {}
            c_data['picture'] = None
            if not char.find('meta', {'itemprop': 'image'}) is None:
                c_data['picture'] = char.find('meta', {'itemprop': 'image'}).get_attribute_list('content')[0]
            c_data['name'] = char.find('span', {'class': 'name-ru'}).text
            res['main_characters'].append(c_data)
        
        r1 = soup.find('div', {'class': 'two-videos'})
        if not r1 is None:
            if not r1.find('div', {'class': 'c-screenshots'}) is None:
                for img in r1.find_all('a', {'class': 'c-screenshot'}):
                    res['screenshots'].append(img.get_attribute_list('href')[0])
            if not r1.find('div', {'class': 'c-videos'}) is None:
                for vid in r1.find_all('div', {'class': 'c-video'}):
                    c_data = {}
                    c_data['link'] = vid.find('a').get_attribute_list('href')[0]
                    c_data['name'] = vid.find('span', {'class': 'name'}).text
                    res['videos'].append(c_data)

        r1 = soup.find('div', {'class': 'block'})
        if not r1 is None:
            for sim in r1.find_all('article'):
                c_data = {}
                c_data['picture'] = None
                if not sim.find('meta', {'itemprop': 'image'}) is None:
                    c_data['picture'] = sim.find('meta', {'itemprop': 'image'}).get_attribute_list('content')[0]
                c_data['name'] = sim.find('span', {'class': 'name-ru'}).text
                c_data['link'] = sim.find('div').get_attribute_list('data-href')[0]
                res['similar'].append(c_data)
        return res
    
    async def get_anime_list(self, status: list[str] = [], anime_type: list[str] = [], rating: str | None = None, genres: list[str] = [], start_page: int = 1, page_limit: int = 3, sort_by: str = 'rating') -> list:
        """
        Получить список аниме по фильтрам

        :status: текущий статус аниме (список нужных) (Доступно: ongoing, anons, released, latest) (По умолчанию пусто - любой)
        :anime_type: Тип аниме (список нужных) (Доступно: tv (тв сериал), movie (фильм), ova, ona, special (спецвыпуски), tv_special (тв спецвыпуск), music (клип), pv (проморолик), cm (реклама)) (По умолчанию пусто - любой)
        :rating: Возрастной рейтинг (Один вариант) (Доступно: g (нет возрастного ограничения), pg (рекомендуется присутствие родителей), pg_13 (детям до 13 просмотр не желателен), r (Лицам до 17 лет обязательно присутствие взрослого), r_plus (Лицам до 17 лет просмотр запрещен)) (По умолчанию пусто - любой)
        :genres: Жанры (список нужных) (Доступно: 
            "1-Action": "Экшен",
            "2-Adventure": "Приключения",
            "3-Racing": "Гонки",
            "4-Comedy": "Комедия",
            "5-Avant-Garde": "Авангард",
            "6-Mythology": "Мифология",
            "7-Mystery": "Тайна",
            "8-Drama": "Драма",
            "9-Ecchi": "Этти",
            "10-Fantasy": "Фэнтези",
            "11-Strategy-Game": "Стратегические игры",
            "13-Historical": "Исторический",
            "14-Horror": "Ужасы",
            "15-Kids": "Детское",
            "17-Martial-Arts": "Боевые искусства",
            "18-Mecha": "Меха",
            "19-Music": "Музыка",
            "20-Parody": "Пародия",
            "21-Samurai": "Самураи",
            "22-Romance": "Романтика",
            "23-School": "Школа",
            "24-Sci-Fi": "Фантастика",
            "25-Shoujo": "Сёдзё",
            "27-Shounen": "Сёнен",
            "29-Space": "Космос",
            "30-Sports": "Спорт",
            "31-Super-Power": "Супер сила",
            "32-Vampire": "Вампиры",
            "35-Harem": "Гарем",
            "36-Slice-of-Life": "Повседневность",
            "37-Supernatural": "Сверхъестественное",
            "38-Military": "Военное",
            "39-Detective": "Детектив",
            "40-Psychological": "Психологическое",
            "42-Seinen": "Сэйнэн",
            "43-Josei": "Дзёсей",
            "102-Team-Sports": "Командный спорт",
            "103-Video-Game": "Видеоигры",
            "104-Adult-Cast": "Взрослые персонажи",
            "105-Gore": "Жестокость",
            "106-Reincarnation": "Реинкарнация",
            "107-Love-Polygon": "Любовный многоугольник",
            "108-Visual-Arts": "Изобразительное искусство",
            "111-Time-Travel": "Путешествие во времени",
            "112-Gag-Humor": "Гэг-юмор",
            "114-Award-Winning": "Удостоено наград",
            "117-Suspense": "Триллер",
            "118-Combat-Sports": "Спортивные единоборства",
            "119-CGDCT": "CGDCT",
            "124-Mahou-Shoujo": "Махо-сёдзё",
            "125-Reverse-Harem": "Реверс-гарем",
            "130-Isekai": "Исэкай",
            "131-Delinquents": "Хулиганы",
            "134-Childcare": "Забота о детях",
            "135-Magical-Sex-Shift": "Магическая смена пола",
            "136-Showbiz": "Шоу-бизнес",
            "137-Otaku-Culture": "Культура отаку",
            "138-Organized-Crime": "Организованная преступность",
            "139-Workplace": "Работа",
            "140-Iyashikei": "Иясикэй",
            "141-Survival": "Выживание",
            "142-Performing-Arts": "Исполнительское искусство",
            "143-Anthropomorphic": "Антропоморфизм",
            "144-Crossdressing": "Кроссдрессинг",
            "145-Idols-(Female)": "Идолы (Жен.)",
            "146-High-Stakes-Game": "Игра с высокими ставками",
            "147-Medical": "Медицина",
            "148-Pets": "Питомцы",
            "149-Educational": "Образовательное",
            "150-Idols-(Male)": "Идолы (Муж.)",
            "151-Romantic-Subtext": "Романтический подтекст",
            "543-Gourmet": "Гурман"
            )
            (По умолчанию пусто - любой)
        :start_page: Начальная страница для поиска (По умолчанию 1)
        :page_limit: ограничение по количеству страниц для парсинга (По умолчанию 3)
        :sort_by: как сортировать выдачу (Доступно: rating (рейтинг), popularity (популярность), name (по алфавиту), aired_on (по дате выхода), ranked_random (случайно), id_desc (по id)) (По умолчанию: rating)

        Возвращает список словарей вида:
        [
            {
                'original_title': 'Оригинальное название (на английском)',
                'poster': 'Ссылка на картинку-постер',
                'shikimori_id': 'id шикимори',
                'title': 'Название на русском',
                'type': 'Тип аниме (TV Сериал, ONA, ...)',
                'url': 'Ссылка на страницу аниме',
                'year': 'год выхода аниме'
            },
            ...
        ]
        """
        if page_limit <= 0:
            return []

        status = [st for st in status if st in ['ongoing', 'anons', 'released', 'latest']] # Убираем все неизвестные варианты чтобы ничего не сломалось
        anime_type = [st for st in anime_type if st in ['tv', 'movie', 'ova', 'ona', 'special', 'tv_special', 'music', 'pv', 'cm']] # Убираем все неизвестные варианты чтобы ничего не сломалось
        if sort_by not in ['rating', 'popularity', 'name', 'aired_on', 'ranked_random', 'id_desc']:
            sort_by = 'rating' # По умолчанию сортировка по рейтингу
        genres = [st for st in genres if st in self.genres_list]

        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:130.0) Gecko/20100101 Firefox/130.0',
            'Accept': 'application/json, text/plain, */*',
            'Referer': f'https://{self._dmn}/animes/status/ongoing',
            'X-Requested-With': 'XMLHttpRequest',
        }

        search_url = f'https://{self._dmn}/animes'
        if len(anime_type) > 0:
            search_url += f'/kind/{",".join(anime_type)}'
        if len(status) > 0:
            search_url += f'/status/{",".join(status)}'
        if len(genres) > 0:
            search_url += f'/genre/{",".join(genres)}'
        if rating != None and rating not in ['g', 'pg' 'pg_13', 'r', 'r_plus']:
            rating = None # Проверяем что введены только доступные (для рейтинга rx требуется аккаунт или пользуйтесь функцией deep_search)

        res = []
        i = start_page
        total_pages = start_page+1 # (После первого запроса обновится)
        while i < start_page+page_limit and i <= total_pages:
            response = await self.requests.get(f'{search_url}/page/{i}.json?order={sort_by}{f"&rating={rating}" if rating != None else ""}', headers=headers)
            if response.status_code == 429:
                raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
            elif response.status_code == 520:
                raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
            elif response.status_code != 200:
                raise errors.ServiceError('Произошла непредвиденная ошибка при получении данных об онгоингах. Ожидался статус ответа 200. Получен: ', response.status_code)
            try:
                data = response.json()
            except json.JSONDecodeError:
                raise errors.UnexpectedBehavior('Ошибка парсинга json при получении данных об онгоингах')
            else:
                total_pages = data['pages_count']
                soup = Soup(data['content'], 'lxml') if self.USE_LXML else Soup(data['content'], 'html.parser')
                articles = soup.find_all('article')
                if len(articles) == 0 and i > 1:
                    return res # В случае если страница есть но данных на ней нет
                elif len(articles) == 0 and i == 1: # Если на первой странице нет данных
                    raise errors.NoResults('Данные по онгоингам не найдены')
                for art in articles:
                    res.append(self._get_anime_info_from_article(art))
            i += 1
        return res
    
    def _get_anime_info_from_article(self, article: Soup) -> dict:
        """
        Используется для получения данных из article
        """
        c_data = {}
        c_data['shikimori_id'] = article.get_attribute_list('id')[0]
        link = article.find('a')
        c_data['url'] = link.get_attribute_list('href')[0]
        try:
            pict = article.find('picture').find('img')
            c_data['poster'] = pict.get_attribute_list('srcset')[0][:-3]
        except:
            c_data['poster'] = None
        try:
            c_data['original_title'] = article.find('span', {'class': 'name-en'}).text
            c_data['title'] = article.find('span', {'class': 'name-ru'}).text
        except: # Если не найдено по отдельности, то скорее всего есть только один вариант
            try:
                c_data['original_title'] = article.find('span', {'class': 'title'}).text
                c_data['title'] = c_data['original_title']
            except:
                c_data['original_title'] = None
                c_data['title'] = None
        c_data['type'] = None
        c_data['year'] = None
        try:
            misc = article.find('span', {'class': 'misc'}).find_all('span')
            for m in misc:
                if m.text.isdigit():
                    c_data['year'] = m.text
                else:
                    c_data['type'] = m.text
        except:
            pass
        return c_data

    async def link_by_id(self, shikimori_id: str) -> str:
        """
        Получить ссылку на страницу аниме по shikimori_id.

        :shikimori_id: id шикимори
        """
        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
        }
        response = await self.requests.get(f'https://{self._dmn}/animes/{shikimori_id}', headers=headers)
        if response.status_code == 404:
            soup = Soup(response.text, 'lxml') if self.USE_LXML else Soup(response.text, 'html.parser')
            actual_code = soup.find('p', {'class': 'error-404'}).text
            if actual_code == '404':
                raise errors.NoResults(f'Страница аниме с shikimori_id "{shikimori_id}" не найдена.')
            elif actual_code == '429':
                raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
            elif actual_code == '302':
                return soup.find('a').get_attribute_list('href')[0]
            else:
                raise errors.UnexpectedBehavior(f'Непредвиденная ошибка при попытке нахождения страницы по id ({shikimori_id}). Ожидались коды: "404", "302", "200". Обнаружен: "{actual_code}"')
        elif response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code == 200 or response.status_code == 302:
            return response.url
        else:
            raise errors.UnexpectedBehavior(f'Непредвиденная ошибка при попытке нахождения страницы по id ({shikimori_id}). Ожидались коды: "404", "302", "200". Обнаружен: "{response.status_code}"')

    def id_by_link(self, shikimori_link: str) -> str:
        """
        Возвращает shikimori_id из ссылки на страницу аниме.

        :shikimori_link: Ссылка на страницу шикимори. (прим: https://shikimori.one/animes/53446-tondemo-skill-de-isekai-hourou-meshi)
        """
        sid = shikimori_link[shikimori_link.rfind('/')+1:shikimori_link.find('-')]
        # Убираем все буквы в id, потому что это шикимори зачем-то их туда начал добавлять
        nsid = ''
        for x in sid:
            if x.isdigit():
                nsid += x
        return nsid
    
    async def deep_search(self, title: str, search_parameters: dict = {}, return_parameters: list = ['id', 'name', 'russian', 'genres { id name russian kind}', 'status', 'url']) -> list:
        """
        Использует playground для graphql. Эмуляция api без необходимости использовать ключ и без ограничений на количество запросов (предположительно).
        ## ! Возможна блокировка ip, используйте на свой страх и риск !
        
        Вы можете прочитать пример запроса в файле: SHIKI_API.md

        :title: Название аниме
        :search_parameters: Параметры поиска. Конкретнее читайте в документации. 
            Пример параметров: {'limit': 10, 'kind': 'tv', 'status': 'released'}
        :return_parameters: Какие данные возвращать. Примеры параметров можно посмотреть тут: https://shikimori.one/api/doc/graphql.
            Пример параметров: ['id', 'name', 'russian', 'genres { id name russian kind }', 'status', 'url']
            Вы можете прочитать пример запроса в файле: SHIKI_API.md
        
        Возвращает список словарей.
        """
        query = "{\n"
        search_query = f'animes(search: \"{title}\"'
        for parameter, value in search_parameters.items():
            if type(value) == str:
                val = f'"{value}"'
            elif type(value) == bool:
                val = 'true' if value else 'false'
            else:
                val = value
            search_query += f', {parameter}: {val}'
        search_query += ")"
        query += search_query
        return_query = '{\n'
        for param in return_parameters:
            return_query += param + '\n'
        return_query += '}'
        query += return_query + '\n}'

        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
            'Accept': '*/*',
            'Referer': f'https://{self._dmn}/api/doc/graphql',
            'content-type': 'application/json',
            'Origin': f'https://{self._dmn}',
        }

        json_data = {
            'operationName': None,
            'variables': {},
            'query': query
        }
        
        response = await self.requests.post(f'https://{self._dmn}/api/graphql', headers=headers, json=json_data)
        if response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.json()
        if 'errors' in response.keys():
            raise errors.PostArgumentsError(f"Ошибка запроса. Ошибка: {response['errors'][0]['message']}")
        response = response['data']
        if 'animes' in response.keys():
            return response['animes']
        else:
            return []
        
    async def deep_anime_info(self, shikimori_id: str, return_parameters: list = ['id', 'name', 'russian', 'genres { id name russian kind}', 'status', 'url']) -> dict:
        """
        Использует playground для graphql. Эмуляция api без необходимости использовать ключ и без ограничений на количество запросов (предположительно).
        ## ! Возможна блокировка ip, используйте на свой страх и риск !

        Вы можете прочитать пример запроса в файле: SHIKI_API.md

        :shikimori_id: id шикимори
        :return_parameters: Какие данные возвращать. Примеры параметров можно посмотреть тут: https://shikimori.one/api/doc/graphql.
            Пример параметров: ['id', 'name', 'russian', 'genres { id name russian kind }', 'status', 'url']
            Вы можете прочитать пример запроса в файле: SHIKI_API.md
        
        Возвращает словарь.
        """
        query = "{\n"
        search_query = f'animes(ids: \"{shikimori_id}\")'
        query += search_query
        return_query = '{\n'
        for param in return_parameters:
            return_query += param + '\n'
        return_query += '}'
        query += return_query + '\n}'

        headers = {
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:128.0) Gecko/20100101 Firefox/128.0',
            'Accept': '*/*',
            'Referer': f'https://{self._dmn}/api/doc/graphql',
            'content-type': 'application/json',
            'Origin': f'https://{self._dmn}',
        }

        json_data = {
            'operationName': None,
            'variables': {},
            'query': query
        }
        
        response = await self.requests.post(f'https://{self._dmn}/api/graphql', headers=headers, json=json_data)
        if response.status_code == 429:
            raise errors.TooManyRequests(f'Сервер вернул код 429 для обозначения что запросы выполняются слишком часто.')
        elif response.status_code == 520:
            raise errors.ServiceIsOverloaded("Сервер вернул статус ответа 520, что означает что он перегружен и не может ответить сразу.")
        elif response.status_code != 200:
            raise errors.ServiceError(f'Сервер не вернул ожидаемый код 200. Код: "{response.status_code}"')
        response = response.json()
        if 'errors' in response.keys():
            raise errors.PostArgumentsError(f"Ошибка запроса. Ошибка: {response['errors'][0]['message']}")
        response = response['data']
        if 'animes' in response.keys():
            if len(response['animes']) > 0:
                return response['animes'][0]
            else:
                raise errors.NoResults(f'Нет данных по shikimori_id "{shikimori_id}"')
        else:
            return {}

"""
Пример запроса для deep_search:

{
  # look for more query params in the documentation
  animes(
    
    // параметры запроса
    // параметр: тип значения - описание параметра
    search: str, - Название аниме
    limit: int, - Максимальное количество выдач (макс: 50)
    rating: str, - Возрастной рейтинг
        // "g" – G - Все возрасты
        // "pg" – PG - Дети
        // "pg_13" – PG-13 - Подростки от 13 лет
        // "r" – R - рекомендовано 17+ (насилие и ненормативная лексика)
        // "r_plus" – R+ - Умеренная нагота (также может содержать насилие и ненормативную лексику)
        // "rx" – Rx - Hentai (экстремальный сексуальный контент или нагота)
    page: int - номер страницы (значение между 1 и 100000)
    order: str - сортировать
        // "id" – по id
        // "ranked" – по рейтингу
        // "kind" – по типу (TV serial, OVA, ...)
        // "popularity" – по популярности
        // "name" – по алфавиту (названия)
        // "aired_on" – по дате выхода
        // "episodes" – по количеству эпизодов
        // "status" – по статусу выхода (вышел, анонс, онгоинг)
        // "random" – перемешать
    kind: str - тип аниме (Одно из: tv, movie, ova, ona, special, tv_special, music, pv, cm, tv_13, tv_24, tv_48)
    status: str - статус выхода (Одно из: anons, ongoing, released)
    season: str - сезон аниме (Примеры: summer_2017, 2016, 2014_2016, 199x)
    score: int - минимальный рейтинг
    duration: str - продолжительность ("S" - меньше 5 минут, "D" - меньше 30 минут, "F" - больше 30 минут)
    genre: str - жанры (указываются id жанров через запятую)
    genre_v2: str - жанры v2 (указываются id v2 жанров через запятую)
    studio: str - студии (указываются id студий через запятую)
    franchize: str - франшизы (указываются id франшиз через запятую)
    censored: boolean - скрывать 18+ (false - чтобы разрешить поиск hentai, yaoi, yuri)
    ids: str - id аниме (указываются через запятую)
    exclude_ids: str - id аниме которые не будут включены в поиск (указываются через запятую)
    ) {
    
    // Возвращаемые параметры
    // параметр - описание возвращенных данных (тип данных)
    //
    id - id шикимори (str)
    malId - id myanimelist (str)
    name - японское название транслитом на английском (str)
    russian - русское название (str)
    licenseNameRu - название русской лицензии (str)
    english - английское название (str)
    japanese - японское название (str) 
    synonyms - список других названий (list[str])
    kind - тип аниме ('tv', 'ova' и т.д.) (str)
    rating - возрастной рейтинг (str)
    score - оценка на шикимори (float)
    status - статус ('released', 'ongoing', 'anons') (str)
    episodes - количество эпизодов (int)
    episodesAired - количество анонсированных эпизодов (int)
    duration - продолжительность серии в минутах (int)
    airedOn { year month day date } - дата анонса (dict) {"year": год (int), "month": месяц (int), "day": день (int), "date": дата в формате "YYYY-MM-DD" (str)}
    releasedOn { year month day date } - дата релиза (dict) {"year": год (int), "month": месяц (int), "day": день (int), "date": дата в формате "YYYY-MM-DD" (str)}
    url - ссылка на страницу на шикимори (str)
    season - сезон аниме (str) (пример: "winter_2023")

    poster { id originalUrl mainUrl } - постер (dict) {"id": id постера (str), "originalUrl": ссылка на jpg картинку (str), 'mainUrl': ссылка на webp картинку (str)}

    fansubbers - субтитры (list[str])
    fandubbers - озвучка (list[str])
    licensors - лицензиаторы? (list)
    createdAt, - дата создания (str)
    updatedAt, - дата обновления (str)
    nextEpisodeAt, - дата выхода следующего эпизода (str или null)
    isCensored - зацензурено ли (boolean)

    genres { id name russian kind } - жанры (list[dict]) {"id": id жанра (str), "name": английское название жанра (str), "russian": русское название жанра (str), "kind": жанр или тема ("genre" или "theme")}
    studios { id name imageUrl } - студии анимации (list[dict]) {"id": id студии (str), "name": название студии (str), "imageUrl": ссылка на картинку студии (str)}

    externalLinks { - внешние ссылки (list[dict])
      id - id ссылки (str)
      kind - тип ссылки (str) Примеры: ("official_site", "wikipedia", "anime_news_network", "myanimelist", "anime_db", "world_art", "kinopoisk", "twitter")
      url - ссылка (str)
      createdAt - дата создания (str)
      updatedAt - дата обновления (str)
    }

    personRoles { - роли-люди (продюсеры, аниматоры, звуковики и т.п.) (list[dict])
      id - id чего-то (str)
      rolesRu - список ролей на русском (list[str])
      rolesEn - список ролей на английском (list[str])
      person { - информация о человеке (dict)
        id - id человека (str)
        name - имя человека (str)
        poster { - данные о фотографии (dict или null)
            id - id фотографии (str)
            mainUrl - ссылка на фотографию webp (str)
            mainAltUrl - ссылка на фотографию jpeg (str)
        } 
      }
    }
    characterRoles { - персонаж-роль (list[dict])
      id - id персонажа (str)
      rolesRu - роли на русском (list[str])
      rolesEn - роли на английском (list[str])
      character { - данные о персонаже (dict)
        id - id персонажа (str)
        name - имя персонажа на английском (str)
        russian - имя персонажа на русском (str)
        poster { - картинка (dict)
            id  - id картинки (str)
            mainUrl - ссылка на webp (str)
            mainAltUrl - ссылка на jpeg (str)
        } 
      }
    }

    related { - связанные (list[dict])
      id - id чего-то (str)
      anime { - связь с аниме (если есть, иначе None) (dict)
        id - id аниме (str)
        name - транслит японского названия на английском (str)
        russian - название на русском (str)
        english - название на английском (str)
        url - ссылка на страницу (str)
      }
      manga { - связь с мангой (если есть, иначе None) (dict)
        id - id манги (str)
        name - транслит японского названия на английском (str)
        russian - название на русском (str)
        english - название на английском (str)
        url - ссылка на страницу (str)
      } // Скорее всего есть и ранобэ
      relationKind - тип связи (str) ("adaptation" - адаптация, "sequel" - продолжение и так далее)
      relationText - тип связи на русском (str)
    }

    videos { - связанные видео (предпросмотры серий) (list[dict])
        id - id чего-то (str)
        url - ссылка на видео (чаще всего ютуб) (str)
        name - имя видео? (str)
        kind - тип видео (episode_preview - трейлер эпизода и т.п.) (str)
        playerUrl - ссылка на плеер (embed ютуба) (str)
        imageUrl - ссылка на обложку видео (тоже ютуб) (str)
    }
    screenshots { - скриншоты (list[dict]) 
        id - id скриншота (str)
        originalUrl - ссылка на оригинальную картинку (str)
        x166Url - ссылка на очень маленькую картинку (str)
        x332Url - ссылка на маленькую картинку (str)
    }

    scoresStats { - статистика оценок (list[dict])
        score - оценка (int)
        count - количество людей поставивших эту оценку (int)
    }
    statusesStats { - статистика статусов просмотра (заброшено, просмотрено, запланировано и т.п.) (list[dict])
        status - статус на английском (str)
        count - количество людей с данным статусом аниме (int)
    }

    description - описание аниме (просто текст) (str)
    descriptionHtml - описание аниме с гиперссылками и html версткой (ыек)
    descriptionSource - откуда взяли описание (если есть, иначе None)
  }
}
"""


import requests
try:
    import lxml
except ImportError:
    LXML_WORKS = False
else:
    LXML_WORKS = True
import json
from bs4 import BeautifulSoup as Soup
from base64 import b64decode

try:
    from . import errors # Импорт если библиотека установлена
except ImportError:
    import errors # Импорт если библиотека не установлена и файл лежит локально

class KodikParser:
    """
    Парсер для плеера Kodik
    """
    def __init__(self, token: str|None = None, use_lxml: bool = False) -> None:
        """
        :token: токен kodik для поиска по его базе. Если не указан будет произведена попытка автоматического получения токена
        :use_lxml: использовать lxml парсер (в некоторых случаях lxml может не работать)
        """
        if token is None:
            try:
                token = self.get_token()
            except Exception as ex:
                raise errors.ServiceError(f'Произошла ошибка при попытке автоматического получения токена kodik. Ошибка: {ex}')
        self.TOKEN = token
        if not LXML_WORKS and use_lxml:
            raise ImportWarning('Параметр use_lxml установлен в true, однако при попытке импорта lxml произошла ошибка')
        self.USE_LXML = use_lxml

    def api_request(self, endpoint: str, filters: dict = {}, parameters: dict = {}) -> dict:
        """
        Запрос к апи кодика по указанным фильтрам и параметрам. Требует токен.
        Более полная документация в файле KODIK_API.md (см. репозиторий https://github.com/YaNesyTortiK/Kodik-Download-Watch)
        
        :endpoint: Эндпоинт для запроса (search, list, translations)
        :filters: Фильтры (по умолчанию {})
        :parameters: Дополнительные параметры (по умолчанию {})

        Возвращает неизмененный ответ сервера.
        """
        if endpoint not in ['search', 'list', 'translations']:
            raise errors.PostArgumentsError(f'Неизвестный эндпоинт. Ожидался один из "search", "list", "translations". Получен: "{endpoint}"')
        if self.TOKEN is None:
            raise errors.TokenError('Токен kodik не указан')
        
        payload = {"token": self.TOKEN}
        for item, val in filters.items():
            payload[item] = val
        for item, val in parameters.items():
            payload[item] = val

        url = f"https://kodikapi.com/{endpoint}"
        data = requests.post(url, data=payload)
        if data.status_code != 200:
            raise errors.ServiceError(f'Произошла ошибка при запросе к kodik api. Ожидался код "200", получен: "{data.status_code}"')
        data = data.json()
        
        if 'error' in data.keys() and data['error'] == 'Отсутствует или неверный токен':
            raise errors.TokenError('Отсутствует или неверный токен')
        elif 'error' in data.keys():
            raise errors.ServiceError(data['error'])
        if data['total'] == 0:
            raise errors.NoResults('Сервер вернул ответ с пустым списком результатов.')
        return data

    def base_search(self, title: str, limit: int = 50, include_material_data: bool = True, anime_status: str|None = None, strict: bool = False) -> dict:
        """
        ### Для использования требуется токен kodik
        Прямой запрос к базе кодика без дополнительных преобразований

        :title: Название аниме / фильма / сериала
        :limit: Верхнее ограничение количества ответов
        :include_material_data: Добавлять в ответ дополнительные данные о сериале
        :anime_status: Статус выхода аниме (доступно: released, ongoing, None - если ищется не аниме или любой статус)
        :strict: Исключение названий далеких от оригинального

        Возвращает словарь следующего вида (на запрос 'Наруто'):
        Некоторые параметры могут отличаться в зависимости от типа/состояния контента
        {
            "time": "18ms",
            "total": 50,
            "results": [
                {
                    "id": "movie-20609",
                    "type": "anime",
                    "link": "//kodik.info/video/20609/e8fd5bc1190b7eb1ee1a3e1c3aec5f62/720p",
                    "title": "Наруто 4",
                    "title_orig": "Gekijô-ban Naruto shippûden",
                    "other_title": "Наруто (фильм четвёртый) - Смерть Наруто / Наруто: Ураганные Хроники - Адепты Тёмного царства",
                    "translation": {
                        "id": 767,
                        "title": "SHIZA Project",
                        "type": "voice"
                    },
                    "year": 2007,
                    "kinopoisk_id": "283418",
                    "imdb_id": "tt0988982",
                    "worldart_link": "http://www.world-art.ru/animation/animation.php?id=6476",
                    "shikimori_id": "2472",
                    "quality": "BDRip 720p",
                    "camrip": false,
                    "lgbt": false,
                    "blocked_countries": [],
                    "created_at": "2018-01-21T13:18:37Z",
                    "updated_at": "2019-11-16T22:13:46Z",
                    "screenshots": [
                        "https://i.kodik.biz/screenshots/video/20609/1.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/2.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/3.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/4.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/5.jpg"
                    ]
                },
                ...
            ],
        }
        """
        try:
            payload = {
                "title": title + ' ' if strict else title, # Хз почему, но если не добавить этот пробел, результатов не будет. \_(=-=)_/
                "limit": limit,
                "with_material_data": 'true' if include_material_data else 'false',
                "strict": 'true' if strict else 'false'
            }
            if anime_status in ['released', 'ongoing']:
                payload['anime_status'] = anime_status
            data = self.api_request(
                'search',
                payload
            )
        except errors.NoResults:
            raise errors.NoResults(f'По запросу "{title}" ничего не найдено')
        return data
    
    def base_search_by_id(self, id: str, id_type: str, limit: int = 50, include_material_data: bool = True) -> dict:
        """
        ### Для использования требуется токен kodik
        Прямой запрос к базе кодика без дополнительных преобразований

        :id: id аниме на одном из сайтов
        :id_type: с какого сайта id (Поддерживается: shikimori, kinopoisk, imdb)
        :limit: Верхнее ограничение количества ответов
        :include_material_data: Добавлять в ответ дополнительные данные о сериале

        Возвращает словарь следующего вида (на запрос 'Наруто'):
        Некоторые параметры могут отличаться в зависимости от типа/состояния контента
        {
            "time": "18ms",
            "total": 50,
            "results": [
                {
                    "id": "movie-20609",
                    "type": "anime",
                    "link": "//kodik.info/video/20609/e8fd5bc1190b7eb1ee1a3e1c3aec5f62/720p",
                    "title": "Наруто 4",
                    "title_orig": "Gekijô-ban Naruto shippûden",
                    "other_title": "Наруто (фильм четвёртый) - Смерть Наруто / Наруто: Ураганные Хроники - Адепты Тёмного царства",
                    "translation": {
                        "id": 767,
                        "title": "SHIZA Project",
                        "type": "voice"
                    },
                    "year": 2007,
                    "kinopoisk_id": "283418",
                    "imdb_id": "tt0988982",
                    "worldart_link": "http://www.world-art.ru/animation/animation.php?id=6476",
                    "shikimori_id": "2472",
                    "quality": "BDRip 720p",
                    "camrip": false,
                    "lgbt": false,
                    "blocked_countries": [],
                    "created_at": "2018-01-21T13:18:37Z",
                    "updated_at": "2019-11-16T22:13:46Z",
                    "screenshots": [
                        "https://i.kodik.biz/screenshots/video/20609/1.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/2.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/3.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/4.jpg",
                        "https://i.kodik.biz/screenshots/video/20609/5.jpg"
                    ]
                },
                ...
            ],
        }
        """
        if id_type not in ['shikimori', 'kinopoisk', 'imdb']:
            raise errors.PostArgumentsError(f'Поддерживаются только id shikimori, kinopoisk, imdb. Получено: {id_type}')
        try:
            data = self.api_request(
                'search',
                {
                    f"{id_type}_id": id,
                    "limit": limit,
                    "with_material_data": 'true' if include_material_data else 'false'
                }
            )
        except errors.NoResults:
            raise errors.NoResults(f'По id {id_type} "{id}" ничего не найдено')
        return data
    
    def get_list(self, limit_per_page: int = 50, pages_to_parse: int = 1, include_material_data: bool = True, anime_status: str|None = None, only_anime: bool = False, start_from: str|None = None) -> tuple[list[dict],str]:
        """
        Получение случайного списка аниме от кодика (скорее всего это будут онгоинги)

        :limit_per_page: Ограничение на количество результатов на запрос(страницу), не все элементы в списке будут аниме (по умолчанию 50)
        :pages_to_parse: Ограничение на количество страниц для обработки (каждая страница - отдельный запрос) (по умолчанию 1)
        :include_material_data: Добавление дополнительных данных (необязательно, по умолчанию True)
        :anime_status: Статус выхода аниме (доступно: released, ongoing, None - если ищется не аниме или любой статус)
        :only_anime: Возвращать только варианты аниме (тип anime или anime-serial) (по умолчанию False)
        :start_from: Поиск следующих страниц по заданному id (id возвращается вторым элементом кортежа) (по умолчанию None)

        Возвращает кортеж из списка словарей и id страницы:
        (
            [
            {
                "title": "Название",
                "type": "тип мультимедиа (anime, film, ...)",
                "year": "Год выпуска фильма",
                "screenshots": [
                    "ссылки на скриншоты"
                ],
                "shikimori_id": "Id шикимори, если нет - None",
                "kinopoisk_id": "Id кинопоиска, если нет - None",
                "imdb_id": "Id imdb, если нет - None",
                "worldart_link": "ссылка на worldart, если нет - None",
                "additional_data": {
                    "Здесь будут находится все остальные данные выданные кодиком, не связанные с отдельным переводом"
                },
                "material_data": { 
                    "Здесь будут все данные о сериале имеющиеся у кодика. (None если указан параметр include_material_data=False)
                    В том числе оценки на шикимори, статус выхода, даты анонсов, выхода, все возможные названия, жанры, студии и многое другое."
                },
                "link": "ссылка на kodik.info (Пример: //kodik.info/video/20609/e8fd5bc1190b7eb1ee1a3e1c3aec5f62/720p)"
            },
            ],
            "next_page_id": "id следующей страницы (для последовательного парсинга нескольких страниц) (может быть None, если след. страниц нет)"
        )
        """
        results = []
        next_page = start_from
        payload = {
            "token": self.TOKEN,
            "limit": limit_per_page,
            "with_material_data": 'true' if include_material_data else 'false'
        }
        if anime_status in ['released', 'ongoing']:
            payload['anime_status'] = anime_status
        for _ in range(pages_to_parse):
            if next_page != None:
                payload['next'] = next_page
            try:
                data = self.api_request(
                    'list',
                    payload
                )
            except errors.NoResults:
                data = {'results': []}
            if 'next_page' in data.keys():
                next_page = data['next_page'][data['next_page'].rfind('=')+1:]
            else:
                next_page = None
            results += data['results']
        return (self._prettify_data(results, only_anime=only_anime), next_page)
    
    def search(self, title: str, limit: int|None = None, include_material_data: bool = True, anime_status: str|None = None, strict: bool = False, only_anime: bool = False) -> list:
        """
        ### Для использования требуется токен kodik
        Получение только самых основных данных о сериале.
        Для получения всех данных воспользуйтесь функцией base_search

        :title: Название аниме / фильма / сериала
        :limit: Верхнее ограничение количества ответов для base_search (необязательно)
        :include_material_data: Добавление дополнительных данных (необязательно, по умолчанию True)
        :anime_status: Статус выхода аниме (доступно: released, ongoing, None - если ищется не аниме или любой статус)
        :strict: Исключение названий далеких от оригинального
        :only_anime: Возвращать только варианты аниме (тип anime или anime-serial) (по умолчанию False)

        Возвращает список словарей в следующем виде:
        [
        {
            "title": "Название",
            "type": "тип мультимедиа (anime, film, ...)",
            "year": "Год выпуска фильма",
            "screenshots": [
                "ссылки на скриншоты"
            ],
            "shikimori_id": "Id шикимори, если нет - None",
            "kinopoisk_id": "Id кинопоиска, если нет - None",
            "imdb_id": "Id imdb, если нет - None",
            "worldart_link": "ссылка на worldart, если нет - None",
            "additional_data": {
                "Здесь будут находится все остальные данные выданные кодиком, не связанные с отдельным переводом"
            },
            "material_data": { 
                "Здесь будут все данные о сериале имеющиеся у кодика. (None если указан параметр include_material_data=False)
                В том числе оценки на шикимори, статус выхода, даты анонсов, выхода, все возможные названия, жанры, студии и многое другое."
            },
            "link": "ссылка на kodik.info (Пример: //kodik.info/video/20609/e8fd5bc1190b7eb1ee1a3e1c3aec5f62/720p)"
        },
        ...
        ]
        """
        if limit is None:
            search_data = self.base_search(title, include_material_data=include_material_data, anime_status=anime_status, strict=strict)
        else:
            search_data = self.base_search(title, limit, include_material_data=include_material_data, anime_status=anime_status, strict=strict)
        return self._prettify_data(search_data['results'], only_anime=only_anime)
    
    def search_by_id(self, id: str, id_type: str, limit: int|None = None) -> list:
        """
        ### Для использования требуется токен kodik
        Получение только самых основных данных о сериале.
        Для получения всех данных воспользуйтесь функцией base_search

        :id: id аниме на одном из сайтов
        :id_type: с какого сайта id (Поддерживается: shikimori, kinopoisk, imdb)
        :limit: Верхнее ограничение количества ответов для base_search (необязательно)

        Возвращает список словарей в следующем виде:
        [
        {
            "title": "Название",
            "type": "тип мультимедиа (anime, film, ...)",
            "year": "Год выпуска фильма",
            "screenshots": [
                "ссылки на скриншоты"
            ],
            "shikimori_id": "Id шикимори, если нет - None",
            "kinopoisk_id": "Id кинопоиска, если нет - None",
            "imdb_id": "Id imdb, если нет - None",
            "worldart_link": "ссылка на worldart, если нет - None",
            "additional_data": {
                "Здесь будут находится все остальные данные выданные кодиком, не связанные с отдельным переводом"
            },
            "material_data": { 
                "Здесь будут все данные о сериале имеющиеся у кодика. (None если указан параметр include_material_data=False)
                В том числе оценки на шикимори, статус выхода, даты анонсов, выхода, все возможные названия, жанры, студии и многое другое."
            },
            "link": "ссылка на kodik.info (Пример: //kodik.info/video/20609/e8fd5bc1190b7eb1ee1a3e1c3aec5f62/720p)"
        },
        ...
        ]
        """
        if type(id) == int:
            id = str(id)
        elif type(id) != str:
            raise ValueError(f'Для id ожидался тип str, получен "{type(id)}"')

        if limit is None:
            search_data = self.base_search_by_id(id, id_type, include_material_data=True)
        else:
            search_data = self.base_search_by_id(id, id_type, limit, include_material_data=True)
        return self._prettify_data(search_data['results'])
    
    def _prettify_data(self, results: list[dict], only_anime: bool = False) -> list[dict]:
        """
        Превращает полученные данные от запроса кодику в удобный вариант словаря

        :results: список словарей (response['results'] в json'е от кодика)
        :only_anime: Возвращать только варианты аниме (тип anime или anime-serial) (по умолчанию False)
        """
        data = []
        added_titles = []
        for res in results:
            if only_anime and res['type'] not in ['anime-serial', 'anime']:
                continue
            if res['title'] not in added_titles:
                additional_data = {}
                for k, i in res.items():
                    if k not in ['title', 'type', 'year', 'screenshots', 'translation',
                                 'shikimori_id', 'kinopoisk_id', 'imdb_id', 'worldart_link',
                                 'id', 'link', 'title_orig', 'other_title', 'created_at', 
                                 'updated_at', 'quality', 'material_data', 'link']:
                        additional_data[k] = i
                
                data.append({
                    'title': res['title'],
                    'title_orig': res['title_orig'],
                    'other_title': res['other_title'] if 'other_title' in res.keys() else None,
                    'type': res['type'],
                    'year': res['year'],
                    'screenshots': res['screenshots'],
                    'shikimori_id': res['shikimori_id'] if 'shikimori_id' in res.keys() else None,
                    'kinopoisk_id': res['kinopoisk_id'] if 'kinopoisk_id' in res.keys() else None,
                    'imdb_id': res['imdb_id'] if 'imdb_id' in res.keys() else None,
                    'worldart_link': res['worldart_link'] if 'worldart_link' in res.keys() else None,
                    'additional_data': additional_data,
                    'material_data': res['material_data'] if 'material_data' in res.keys() else None,
                    'link': res['link']
                })
                added_titles.append(res['title'])
        return data
    
    def translations(self, id: str, id_type: str) -> list:
        """
        ### Для использования требуется токен kodik
        Возвращает список переводов для медиафайла по id.

        :id: id медиа
        :id_type: тип id (возможные: shikimori, kinopoisk, imdb)

        Пример возвращаемого:
        [
            {
                'id': id перевода/субтитров/озвучки
                'type': тип (voice/subtitles)
                'name': Имя автора озвучки/субтитров
            },
            ...
        ]
        """
        return self.get_info(id, id_type)['translations']
    
    def series_count(self, id: str, id_type: str) -> int:
        """
        ### Для использования требуется токен kodik
        Возвращает количество серий для медиафайла по id.

        :id: id медиа
        :id_type: тип id (возможные: shikimori, kinopoisk, imdb)
        """
        return self.get_info(id, id_type)['series_count']

    def _link_to_info(self, id: str, id_type: str, https: bool = True) -> str:
        """
        ### Для использования требуется токен kodik
        Получить ссылку до страницы с данными.

        :id: id медиа
        :id_type: тип id (возможные: shikimori, kinopoisk, imdb)

        Возвращает ссылку
        """
        if self.TOKEN is None:
            raise errors.TokenError('Токен kodik не указан')
        if id_type == "shikimori":
            serv = f"https://kodikapi.com/get-player?title=Player&hasPlayer=false&url=https%3A%2F%2Fkodikdb.com%2Ffind-player%3FshikimoriID%3D{id}&token={self.TOKEN}&shikimoriID={id}"
        elif id_type == "kinopoisk":
            serv = f"https://kodikapi.com/get-player?title=Player&hasPlayer=false&url=https%3A%2F%2Fkodikdb.com%2Ffind-player%3FkinopoiskID%3D{id}&token={self.TOKEN}&kinopoiskID={id}"
        elif id_type == "imdb":
            serv = f"https://kodikapi.com/get-player?title=Player&hasPlayer=false&url=https%3A%2F%2Fkodikdb.com%2Ffind-player%3FkinopoiskID%3D{id}&token={self.TOKEN}&imdbID={id}"
        else:
            raise ValueError("Неизвестный тип id")
        data = requests.get(serv).json()
        if 'error' in data.keys() and data['error'] == 'Отсутствует или неверный токен':
            raise errors.TokenError('Отсутствует или неверный токен')
        elif 'error' in data.keys():
            raise errors.ServiceError(data['error'])
        if not data['found']:
            raise errors.NoResults(f'Нет данных по {id_type} id "{id}"')
        return 'https:'+data['link'] if https else 'http:'+data['link']
    
    def get_info(self, id: str, id_type: str) -> dict:
        """
        ### Для использования требуется токен kodik

        :id: id медиа
        :id_type: тип id (возможные: shikimori, kinopoisk, imdb)
        
        Возвращает данные:
        {
            'series_count': количество серий (если серий нет/фильм значение будет 0)
            'translations': [
                {
                    'id': id перевода/субтитров/озвучки
                    'type': тип (voice/subtitles)
                    'name': Имя автора озвучки/субтитров
                },
                ...
            ]
        }
        """
        if type(id) == int:
            id = str(id)
        elif type(id) != str:
            raise ValueError(f'Для id ожидался тип str, получен "{type(id)}"')
    
        link = self._link_to_info(id, id_type)
        data = requests.get(link).text
        soup = Soup(data, 'lxml') if self.USE_LXML else Soup(data, 'html.parser')
        if self._is_serial(link):
            series_count = len(soup.find("div", {"class": "serial-series-box"}).find("select").find_all("option"))
            try:
                translations_div = soup.find("div", {"class": "serial-translations-box"}).find("select").find_all("option")
            except:
                translations_div = None
            return {
                'series_count': series_count,
                'translations': self._generate_translations_dict(translations_div)
            }
        elif self._is_video(link):
            series_count = 0
            try:
                translations_div = soup.find("div", {"class": "movie-translations-box"}).find("select").find_all("option")
            except AttributeError:
                translations_div = None
            return {
                'series_count': series_count,
                'translations': self._generate_translations_dict(translations_div)
            }
        else:
            raise errors.UnexpectedBehavior('Ссылка на данные не была распознана как ссылка на сериал или фильм')
    
    def _is_serial(self, iframe_url: str) -> bool:
        return True if iframe_url[iframe_url.find(".info/")+6] == "s" else False

    def _is_video(self, iframe_url: str) -> bool:
        return True if iframe_url[iframe_url.find(".info/")+6] == "v" else False
    
    def _generate_translations_dict(self, translations_div: Soup) -> dict:
        if not isinstance(translations_div, Soup) and translations_div != None:
            translations = []
            for translation in translations_div:
                a = {}
                a['id'] = translation['value']
                a['type'] = translation['data-translation-type']
                if a['type'] == 'voice':
                    a['type'] = "Озвучка"
                elif a['type'] == 'subtitles':
                    a['type'] = "Субтитры"
                a['name'] = translation.text
                translations.append(a)
        else:
            translations = [{"id": "0", "type": "Неизвестно", "name": "Неизвестно"}]
        return translations

    def get_link(self, id: str, id_type: str, seria_num: int, 
                 translation_id: str, crypted: bool = False) -> tuple[str, int]:
        """
        ### Для использования требуется токен kodik
        Возвращает ссылку на видео файл.

        :id: id медиа
        :id_type: тип id (возможные: shikimori, kinopoisk, imdb)
        :seria_num: номер серии (если фильм или одно видео, укажите 0)
        :translation_id: id перевода (прим: Anilibria = 610, если неизвестно - 0)
        :crypted: Используется ли шифрование ссылки кодиком. По умолчанию False.

        Возвращает ссылку в стиле:
        //cloud.kodik-storage.com/useruploads/351182fc-e1ac-4521-a9e3-261303e69687/ba18e7c1a8ac055a61b0d2e528f9eb8c:2024062702/
        В конце ссылки нужно добавить качество видео: 720.mp4 / 480.mp4 / 360.mp4
        В начале ссылки нужно добавить: http: / https:

        И максимально возможное качество.
        """
        # Проверка переданных параметров на правильность типа
        if type(id) == int:
            id = str(id)
        elif type(id) != str:
            raise ValueError(f'Для id ожидался тип str, получен "{type(id)}"')
        if type(seria_num) == str and seria_num.isdigit():
            seria_num = int(seria_num)
        elif type(seria_num) != int:
            raise ValueError(
                f'Для seria_num ожидался тип int, получен "{type(seria_num)}"'
            )
        if type(translation_id) == int:
            translation_id = str(translation_id)
        elif type(translation_id) != str:
            raise ValueError(
                f'Для translation_id ожидался тип str, получен "{type(translation_id)}"'
            )

        link = self._link_to_info(id, id_type)
        data = requests.get(link).text
        soup = Soup(data, "lxml") if self.USE_LXML else Soup(data, "html.parser")
        urlParams = data[data.find("urlParams") + 13 :]
        urlParams = json.loads(urlParams[: urlParams.find(";") - 1])
        if (
            translation_id != "0" and seria_num != 0
        ):  # Обычный сериал с известной озвучкой на более чем 1 серию
            container = soup.find("div", {"class": "serial-translations-box"}).find(
                "select"
            )
            media_hash = None
            media_id = None
            for translation in container.find_all("option"):
                if translation.get_attribute_list("data-id")[0] == translation_id:
                    media_hash = translation.get_attribute_list("data-media-hash")[0]
                    media_id = translation.get_attribute_list("data-media-id")[0]
                    break
            url = f"https://kodik.info/serial/{media_id}/{media_hash}/720p?min_age=16&first_url=false&season=1&episode={seria_num}"
            data = requests.get(url).text
            soup = Soup(data, "lxml") if self.USE_LXML else Soup(data, "html.parser")
        elif (
            translation_id != "0" and seria_num == 0
        ):  # Фильм/одна серия с несколькими переводами
            container = soup.find("div", {"class": "movie-translations-box"}).find(
                "select"
            )
            media_hash = None
            media_id = None
            for translation in container.find_all("option"):
                if translation.get_attribute_list("data-id")[0] == translation_id:
                    media_hash = translation.get_attribute_list("data-media-hash")[0]
                    media_id = translation.get_attribute_list("data-media-id")[0]
                    break
            url = f"https://kodik.info/video/{media_id}/{media_hash}/720p?min_age=16&first_url=false&season=1&episode={seria_num}"
            data = requests.get(url).text
            soup = Soup(data, "lxml") if self.USE_LXML else Soup(data, "html.parser")
        script_url = soup.find_all("script")[1].get_attribute_list("src")[0]

        hash_container = soup.find_all("script")[4].text
        video_type = hash_container[hash_container.find(".type = '") + 9 :]
        video_type = video_type[: video_type.find("'")]
        video_hash = hash_container[hash_container.find(".hash = '") + 9 :]
        video_hash = video_hash[: video_hash.find("'")]
        video_id = hash_container[hash_container.find(".id = '") + 7 :]
        video_id = video_id[: video_id.find("'")]

        link_data, max_quality = self._get_link_with_data(
            video_type, video_hash, video_id, urlParams, script_url, crypted
        )

        download_url = str(link_data).replace("https:", "")[:-25]
        return download_url, max_quality

    def _get_link_with_data(self, video_type: str, video_hash: str, video_id: str, 
                            urlParams: dict, script_url: str, crypted: bool = False):
        params = {
            "hash": video_hash,
            "id": video_id,
            "type": video_type,
            'd': urlParams['d'],
            'd_sign': urlParams['d_sign'],
            'pd': urlParams['pd'],
            'pd_sign': urlParams['pd_sign'],
            'ref': '',
            'ref_sign': urlParams['ref_sign'],
            'bad_user': 'true',
            'cdn_is_working': 'true',
        }

        headers = {"Content-Type": "application/x-www-form-urlencoded"}

        post_link = self._get_post_link(script_url)
        data = requests.post(f"https://kodik.info{post_link}", data=params, headers=headers).json()
        data_url = data["links"]["360"][0]["src"]
        url = data_url if not crypted else self._convert(data_url)
        max_quality = max([int(x) for x in data["links"].keys()])

        return url, max_quality

    def _convert_char(self, char: str):
        low = char.islower()
        alph = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        if char.upper() in alph:
            ch = alph[(alph.index(char.upper()) + 13) % len(alph)]
            if low:
                return ch.lower()
            else:
                return ch
        else:
            return char

    def _convert(self, string: str):
        # Декодирование строки со ссылкой
        crypted_url = "".join(map(self._convert_char, list(string)))
        try:
            return b64decode(crypted_url.encode())
        except:
            return str(b64decode(crypted_url.encode() + b"=="))

    def _get_post_link(self, script_url: str):
        data = requests.get("https://kodik.info" + script_url).text
        url = data[data.find("$.ajax") + 30 : data.find("cache:!1") - 3]
        return b64decode(url.encode()).decode()

    @staticmethod
    def get_token() -> str:
        """
        Попытка получения токена.
        Обратите внимание, что эта функция может не работать из-за изменений кодиком ссылок.
        """
        script_url = 'https://kodik-add.com/add-players.min.js?v=2'
        data = requests.get(script_url).text
        token = data[data.find('token=')+7:]
        token = token[:token.find('"')]
        return token



try:
    from . import parser_kodik # Импорт если библиотека установлена
except ImportError:
    import parser_kodik # Импорт если библиотека не установлена и файл лежит локально

try:
    from . import errors # Импорт если библиотека установлена
except ImportError:
    import errors # Импорт если библиотека не установлена и файл лежит локально

try:
    try:
        from .internal_tools import AsyncSession
        async_available = True
    except ImportError:
        from internal_tools import AsyncSession
        async_available = True
except ImportError:
    async_available = False

import requests

class _AnimeKind:
    tv = 'tv'
    tv13 = 'tv13'
    tv24 = 'tv24'
    tv48 = 'tv48'
    movie = 'movie'
    special = 'special'
    ova = 'ova'
    ona = 'ona'
    music = 'music'

    @staticmethod
    def get_list() -> list[str]:
        """
        Возвращает список доступных параметров.
        """
        return [
            _AnimeKind.tv, _AnimeKind.tv13, _AnimeKind.tv24, _AnimeKind.tv48, 
            _AnimeKind.movie, _AnimeKind.special, _AnimeKind.ova, _AnimeKind.ona, _AnimeKind.music
        ]

class _Types:
    foreign_movie = "foreign-movie"
    soviet_cartoon = "soviet-cartoon"
    foreign_cartoon = "foreign-cartoon"
    russian_cartoon = "russian-cartoon"
    anime = "anime"
    russian_movie = "russian-movie"
    cartoon_serial = "cartoon-serial"
    documentary_serial = "documentary-serial"
    russian_serial = "russian-serial"
    foreign_serial = "foreign-serial"
    anime_serial = "anime-serial"
    multi_part_film = "multi-part-film"

    @staticmethod
    def get_list() -> list[str]:
        """
        Возвращает список доступных параметров.
        """
        return [
            _Types.foreign_movie, _Types.soviet_cartoon, _Types.foreign_cartoon, _Types.russian_cartoon, _Types.anime,
            _Types.russian_movie, _Types.cartoon_serial, _Types.documentary_serial, _Types.russian_serial,
            _Types.foreign_serial, _Types.anime_serial, _Types.multi_part_film
        ]

class _MPAArating:
    g = "g" # - Нет возрастных ограничений.
    pg = "pg" # - Рекомендуется присутствие родителей.
    pg_13 = "pg-13" # - Детям до 13 лет просмотр нежелателен.
    r = "r" # - Лицам до 17 лет обязательно присутствие взрослого.
    rx = "rx" # - Хeнтай / Пoрнография

    @staticmethod
    def get_list() -> list[str]:
        """
        Возвращает список доступных параметров.
        """
        return [_MPAArating.g, _MPAArating.pg, _MPAArating.pg_13, _MPAArating.r, _MPAArating.rx]

class _Genres:
    anime = "аниме"
    drama = "драма"
    comedy = "комедия"
    cartoon = "мультфильм"
    fantasy = "фэнтези" 
    action = "боевик"
    adventures = "приключения"
    fantastika = "фантастика" # А в чем разница между фэнтези и фантастикой? даже гугл одинаково переводит 	(ノ ゜Д゜)ノ ︵

    @staticmethod
    def get_list() -> list[str]:
        """
        Возвращает список доступных параметров.
        """
        return [
            _Genres.anime, _Genres.drama, _Genres.comedy, _Genres.cartoon, _Genres.fantasy,
            _Genres.action, _Genres.adventures, _Genres.fantastika
        ]

class _AnimeGenres: 
    military = "Военное"
    drama = "Драма"
    history = "Исторический"
    action = "Экшен"
    adventures = "Приключения"
    senen = "Сёнен"
    fantasy = "Фэнтези"
    comedy = "Комедия"
    martial_arts = "Боевые искусства"
    romance = "Романтика"
    psychological = "Психологическое"
    thriller = "Триллер"
    everyday_life = "Повседневность"
    supernatural = "Сверхъестественное"
    sport = "Спорт"
    school = "Школа"
    music = "Музыка"
    fantastika = "Фантастика"
    samurai = "Самураи"

    @staticmethod
    def get_list() -> list[str]:
        """
        Возвращает список доступных параметров.
        """
        return [
            _AnimeGenres.military, _AnimeGenres.drama, _AnimeGenres.history, _AnimeGenres.action,
            _AnimeGenres.adventures, _AnimeGenres.senen, _AnimeGenres.fantasy, _AnimeGenres.comedy,
            _AnimeGenres.martial_arts, _AnimeGenres.romance, _AnimeGenres.psychological, _AnimeGenres.thriller,
            _AnimeGenres.everyday_life, _AnimeGenres.supernatural, _AnimeGenres.sport, _AnimeGenres.school, 
            _AnimeGenres.music, _AnimeGenres.fantastika, _AnimeGenres.samurai
        ]

class Response:
    class Translation:
        def __init__(self, raw_data: dict):
            self.raw_data = raw_data
            self._keys = raw_data.keys()
            self.id = raw_data['id'] if 'id' in self._keys else None
            self.title = raw_data['title'] if 'title' in self._keys else None
            self.type = raw_data['type'] if 'type' in self._keys else None
            self.is_voice = True if self.type == 'voice' else False

        def __str__(self):
            return str(self.raw_data)
        
        def __repr__(self):
            return str(self.raw_data)

    class Season:
        class Episode:
            def __init__(self, raw_data: dict):
                self.raw_data = raw_data
                if type(raw_data) == dict:
                    self.link = raw_data['link']
                    self.title = raw_data['title'] if 'title' in self.raw_data.keys() else None
                    self.screenshots = raw_data['screenshots'] if 'screenshots' in self.raw_data.keys() else None
                else:
                    self.link = raw_data
                    self.title = None
                    self.screenshots = None

        def __init__(self, raw_data: dict):
            self.raw_data = raw_data
            self.episodes = self.raw_data['episodes'] if 'episodes' in self.raw_data.keys() else None
            self.link = self.raw_data['link'] if 'link' in self.raw_data.keys() else None

        @property
        def episodes(self) -> dict[str, 'Response.Season.Episode']:
            return self._episodes
        
        @episodes.setter
        def episodes(self, value: dict | None):
            if value is None:
                self._episodes = None
            else:
                self._episodes = {}
                for k, v in value.items():
                    self._episodes[k] = Response.Season.Episode(v)

    class MaterialData:
        """
        Поле material_data. (Доступно при выполнении запроса с параметром with_material_data)
        Доступные параметры:

        * title - Название
        * anime_title - Название аниме (если есть, иначе None, обычно совпадает с title)
        * title_en - Английское название (либо транслитерация)
        * other_titles - Список других названий (если есть, иначе None)
        * other_titles_en - Список других названий на английском (если есть, иначе None)
        * other_titles_jp - Список других названий на японском (если есть, иначе None)
        * anime_kind - Тип аниме (если есть, иначе None)
        * all_status - Статус выхода
        * anime_status - Статус выхода, если аниме (обычно совпадает с all_status) (если есть, иначе None)
        * description - Описание
        * anime_description - Описание, если аниме (обычно совпадает с description) (если есть, иначе None)
        * poster_url - Ссылка на постер (если есть, иначе ссылка на картинку по умолчанию)
        * anime_poster_url - Ссылка на аниме постер (если есть, то на шикимори, иначе None)
        * screenshots - Список ссылок на скриншоты
        * duration - Продолжительность в минутах
        * all_genres - Список всех жанров (аниме жанры + обычные жанры)
        * anime_genres - Список аниме жанров (если есть, иначе None)
        * shikimori_rating - Оценка на шикимори (если есть, иначе None)
        * shikimori_votes - Количество оценок на шикимори (если есть, иначе None)
        * kinopoisk_rating - Оценка на кинопоиске (если есть, иначе None)
        * kinopoisk_votes - Количество оценок на кинопоиске (если есть, иначе None)
        * imdb_rating - Оценка на imdb (если есть, иначе None)
        * imdb_votes - Количество оценок на imdb (если есть, иначе None)
        * aired_at - Когда проходила премьера (если есть, иначе None)
        * rating_mpaa - Возрастной рейтинг mpaa
        * minimal_age - Минимальный возраст зрителя
        * episodes_total - Всего эпизодов (если есть, иначе None)
        * episodes_aired - Сколько эпизодов уже показано (если есть, иначе None)
        * year - Год выхода 
        * countries - Страны-производители (список) (если есть, иначе None)
        * genres - Список жанров
        * premiere_world - Дата премьеры в мире (если есть, иначе None)
        * actors - Список актеров (если есть, иначе None)
        * directors - Список режиссеров (если есть, иначе None)
        * writers - Список сценаристов (если есть, иначе None)
        * producers - Список продюсеров (если есть, иначе None)
        * composers - Список композиторов (если есть, иначе None)
        * editors - Список монтажеров (если есть, иначе None)
        * designers - Список дизайнеров (если есть, иначе None)
        * operators - Список операторов (если есть, иначе None)
        * licensed_by - Кем было лицензировано (если есть, иначе None)
        * anime_studios - Студии анимации аниме (если есть, иначе None)
        * released_at - Дата выхода (если есть, иначе None)
        """
        def __init__(self, raw_data: dict):
            self.raw_data = raw_data
            self._keys = self.raw_data.keys()
            self.title = self.raw_data['title'] if 'title' in self._keys else None
            self.anime_title = self.raw_data['anime_title'] if 'anime_title' in self._keys else None
            self.title_en = self.raw_data['title_en'] if 'title_en' in self._keys else None
            self.other_titles = self.raw_data['other_titles'] if 'other_titles' in self._keys else None
            self.other_titles_en = self.raw_data['other_titles_en'] if 'other_titles_en' in self._keys else None
            self.other_titles_jp = self.raw_data['other_titles_jp'] if 'other_titles_jp' in self._keys else None
            self.anime_kind = self.raw_data['anime_kind'] if 'anime_kind' in self._keys else None
            self.all_status = self.raw_data['all_status'] if 'all_status' in self._keys else None
            self.anime_status = self.raw_data['anime_status'] if 'anime_status' in self._keys else None
            self.description = self.raw_data['description'] if 'description' in self._keys else None
            self.anime_description = self.raw_data['anime_description'] if 'anime_description' in self._keys else None
            self.poster_url = self.raw_data['poster_url'] if 'poster_url' in self._keys else None
            self.anime_poster_url = self.raw_data['anime_poster_url'] if 'anime_poster_url' in self._keys else None
            self.screenshots = self.raw_data['screenshots'] if 'screenshots' in self._keys else None
            self.duration = self.raw_data['duration'] if 'duration' in self._keys else None
            self.all_genres = self.raw_data['all_genres'] if 'all_genres' in self._keys else None
            self.anime_genres = self.raw_data['anime_genres'] if 'anime_genres' in self._keys else None
            self.shikimori_rating = self.raw_data['shikimori_rating'] if 'shikimori_rating' in self._keys else None
            self.shikimori_votes = self.raw_data['shikimori_votes'] if 'shikimori_votes' in self._keys else None
            self.kinopoisk_rating = self.raw_data['kinopoisk_rating'] if 'kinopoisk_rating' in self._keys else None
            self.kinopoisk_votes = self.raw_data['kinopoisk_votes'] if 'kinopoisk_votes' in self._keys else None
            self.imdb_rating = self.raw_data['imdb_rating'] if 'imdb_rating' in self._keys else None
            self.imdb_votes = self.raw_data['imdb_votes'] if 'imdb_votes' in self._keys else None
            self.aired_at = self.raw_data['aired_at'] if 'aired_at' in self._keys else None
            self.rating_mpaa = self.raw_data['rating_mpaa'] if 'rating_mpaa' in self._keys else None
            self.minimal_age = self.raw_data['minimal_age'] if 'minimal_age' in self._keys else None
            self.episodes_total = self.raw_data['episodes_total'] if 'episodes_total' in self._keys else None
            self.episodes_aired = self.raw_data['episodes_aired'] if 'episodes_aired' in self._keys else None
            self.year = self.raw_data['year'] if 'year' in self._keys else None
            self.countries = self.raw_data['countries'] if 'countries' in self._keys else None
            self.all_genres = self.raw_data['all_genres'] if 'all_genres' in self._keys else None
            self.genres = self.raw_data['genres'] if 'genres' in self._keys else None
            self.anime_genres = self.raw_data['anime_genres'] if 'anime_genres' in self._keys else None
            self.premiere_world = self.raw_data['premiere_world'] if 'premiere_world' in self._keys else None
            self.actors = self.raw_data['actors'] if 'actors' in self._keys else None
            self.directors = self.raw_data['directors'] if 'directors' in self._keys else None
            self.writers = self.raw_data['writers'] if 'writers' in self._keys else None
            self.producers = self.raw_data['producers'] if 'producers' in self._keys else None
            self.composers = self.raw_data['composers'] if 'composers' in self._keys else None
            self.editors = self.raw_data['editors'] if 'editors' in self._keys else None
            self.designers = self.raw_data['designers'] if 'designers' in self._keys else None
            self.operators = self.raw_data['operators'] if 'operators' in self._keys else None
            self.licensed_by = self.raw_data['licensed_by'] if 'licensed_by' in self._keys else None
            self.anime_studios = self.raw_data['anime_studios'] if 'anime_studios' in self._keys else None
            self.released_at = self.raw_data['released_at'] if 'released_at' in self._keys else None

    class Element:
        """
        Описывает элемент. Доступные параметры:

        * id - Внутренний id элемента на кодике
        * type - Тип элемента
        * link - ссылка на embed (kodik.info/....)
        * title - Название на русском
        * title_orig - Оригинальное название (название на английском) (если есть, иначе None)
        * other_title - Другое название (если есть, иначе None)
        * year - Год выхода
        * last_season - Последний сезон (если есть, иначе None)
        * last_episode - Последний эпизод (если есть, иначе None)
        * episodes_count - Количество эпизодов (если есть, иначе None)
        * kinopoisk_id - id элемента на кинопоиске (если есть, иначе None)
        * shikimori_id - id элемента на шикимори (если есть, иначе None)
        * imdb_id - id элемента на imdb (если есть, иначе None)
        * mdl_id - id элемента на my drama list (если есть, иначе None)
        * quality - Качество
        * camrip - Снято ли с камеры
        * lgbt - Есть ли лгбт сцены
        * blocked_countries - Список заблокированных стран
        * blocked_seasons - Список заблокированных сезонов
        * created_at - Дата создания элемента
        * updated_at - Дата обновления элемента
        * screenshots - Список ссылок на скриншоты
        * translation - Перевод (тип Response.Translation)
        * material_data - Дополнительные параметры (тип Response.MaterialData) (если есть, иначе None)
        * seasons - Словарь сезонов (вид: '{'1': Response.Season}') (если есть, иначе None)
        """
        def __init__(self, raw_data: dict):
            self.raw_data = raw_data
            self._keys = self.raw_data.keys()
            self.id = self.raw_data['id'] if 'id' in self._keys else None
            self.type = self.raw_data['type'] if 'type' in self._keys else None
            self.link = self.raw_data['link'] if 'link' in self._keys else None
            self.title = self.raw_data['title'] if 'title' in self._keys else None
            self.title_orig = self.raw_data['title_orig'] if 'title_orig' in self._keys else None
            self.other_title = self.raw_data['other_title'] if 'other_title' in self._keys else None
            self.year = self.raw_data['year'] if 'year' in self._keys else None
            self.last_season = self.raw_data['last_season'] if 'last_season' in self._keys else None
            self.last_episode = self.raw_data['last_episode'] if 'last_episode' in self._keys else None
            self.episodes_count = self.raw_data['episodes_count'] if 'episodes_count' in self._keys else None
            self.kinopoisk_id = self.raw_data['kinopoisk_id'] if 'kinopoisk_id' in self._keys else None
            self.shikimori_id = self.raw_data['shikimori_id'] if 'shikimori_id' in self._keys else None
            self.imdb_id = self.raw_data['imdb_id'] if 'imdb_id' in self._keys else None
            self.mdl_id = self.raw_data['mdl_id'] if 'mdl_id' in self._keys else None
            self.quality = self.raw_data['quality'] if 'quality' in self._keys else None
            self.camrip = self.raw_data['camrip'] if 'camrip' in self._keys else None
            self.lgbt = self.raw_data['lgbt'] if 'lgbt' in self._keys else None
            self.blocked_countries = self.raw_data['blocked_countries'] if 'blocked_countries' in self._keys else None
            self.blocked_seasons = self.raw_data['blocked_seasons'] if 'blocked_seasons' in self._keys else None
            self.created_at = self.raw_data['created_at'] if 'created_at' in self._keys else None
            self.updated_at = self.raw_data['updated_at'] if 'updated_at' in self._keys else None
            self.screenshots = self.raw_data['screenshots'] if 'screenshots' in self._keys else None
            self.translation = Response.Translation(self.raw_data['translation']) if 'translation' in self._keys else None
            self.material_data  = self.raw_data['material_data'] if 'material_data' in self._keys else None
            self.seasons = self.raw_data['seasons'] if 'seasons' in self._keys else None
        
        @property
        def seasons(self) -> dict[str, 'Response.Season']:
            return self._seasons
        
        @seasons.setter
        def seasons(self, value: dict | None):
            if value is None:
                self._seasons = None
            else:
                self._seasons = {}
                for k, v in value.items():
                    self._seasons[k] = Response.Season(v)
        
        @property
        def material_data(self) -> 'Response.MaterialData':
            return self._material_data
        
        @material_data.setter
        def material_data(self, value: dict | None):
            if value is None:
                self._material_data = None
            else:
                self._material_data = Response.MaterialData(value)

    def __init__(self, raw_data: dict):
        self.raw_data = raw_data
        self.total = self.raw_data['total'] # Гарантируется не менее одного результата т.к. при выполнении api_request происходит проверка на наличие результатов. При их отсутствии выходит NoResults exception
        self.time = self.raw_data['time']
        self.results = self.raw_data['results']

    @property
    def results(self) -> list['Response.Element']:
        return self._results
    
    @results.setter
    def results(self, value: list):
        self._results = [Response.Element(r) for r in value]

class Api:
    class AnimeKind(_AnimeKind):
        """
        Тип аниме по виду. Доступные фильтры:\n
        * tv - ТВ Сериал
        * tv13 - Длительность серии ~13 минут
        * tv24 - Длительность серии ~24 минут
        * tv48 - Длительность серии ~48 минут
        * movie - Фильм
        * special - Спец. выпуск
        * ova
        * ona
        * music - Музыкальный клип / Проморолик
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)\n
        """
        pass

    class Types(_Types):
        """
        Тип мультимедиа. Доступные фильтры:\n
        * foreign-movie - Иностранный фильм
        * soviet-cartoon - Советский мультфильм
        * foreign-cartoon - Иностранный мультфильм
        * russian-cartoon - Русский мультфильм
        * anime - Аниме фильм (либо одна серия)
        * russian-movie - Русский фильм
        * cartoon-serial - мульт-сериал
        * documentary-serial - документальный сериал
        * russian-serial - Русский сериал
        * foreign-serial - Иностранный сериал
        * anime-serial - Аниме сериал (все, что больше 1 серии)
        * multi-part-film - Фильм из нескольких частей
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)
        """
        pass

    class MPAArating(_MPAArating):
        """
        Возрастной рейтинг mpaa. Доступные фильтры:\n
        * g - Нет возрастных ограничений.
        * pg - Рекомендуется присутствие родителей.
        * pg-13 - Детям до 13 лет просмотр нежелателен.
        * r - Лицам до 17 лет обязательно присутствие взрослого.
        * rx - Хeнтай / Пoрнография    
        \nЧисто технически, существует рейтинг r+, но почему-то у меня не получилось отфильтровать по нему поиск (выдавался рейтинг r).
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)
        """
        pass

    class Genres(_Genres):
        """
        Жанры (отдельно от аниме жанров). Доступные фильтры:\n    
        * anime - аниме
        * drama - драма
        * comedy - комедия
        * cartoon - мультфильм
        * fantasy - фэнтези
        * action - боевик
        * adventures - приключения
        * fantastika - фантастика
        \n(Для передачи параметра напрямую, строкой, укажите его на русском, как указано выше)
        """
        pass

    class AnimeGenres(_AnimeGenres):
        """
        Аниме жанры/тематики. Доступные фильтры:\n
        * military - Военное
        * drama - Драма
        * history - Исторический
        * action - Экшен
        * adventures - Приключения
        * senen - Сёнен
        * fantasy - Фэнтези
        * comedy - Комедия
        * martial_arts - Боевые искусства
        * romance - Романтика
        * psychological - Психологическое
        * thriller - Триллер
        * everyday_life - Повседневность
        * supernatural - Сверхъестественное
        * sport - Спорт
        * school - Школа
        * music - Музыка
        * fantastika - Фантастика
        * samurai - Самураи
        \n(Для передачи параметра напрямую, строкой, укажите его на русском, как указано выше)
        """
        pass

    def __init__(self, token: str | None = None, allow_warnings: bool = True, _args: dict = {}, _endpoint: str | None = None):
        """
        :token: Указать кастомный токен для апи кодика. По умолчанию None и токен будет подставляться автоматически.
        :allow_warnings: Разрешить вывод в консоль предупреждений о неправильно/некорректно используемых параметрах. По умолчанию True.
        :_args: Используется для передачи данных между классами.
        :_parser: Используется для передачи объекта парсера между классами. (Чтобы каждый раз не инициализировать заново и не ждать автоматического поиска токена)
        :_endpoint: Указывается конец ссылки для апи (search/list/translations). По умолчанию None, т.к. для каждого эндпоинта есть свои условия, которые предусмотрены в классах Search, List соответственно. 
        """
        if token is None:
            self.token = parser_kodik.KodikParser.get_token()
        else:
            self.token = token
        if not async_available and allow_warnings:
            print('При попытке импорта внутренних инструментов библиотеки anime-parsers-ru произошла ошибка и AsyncSession не доступен. Вы не сможете использовать асинхронные функции.')
        self._args = _args
        self.allow_warnings = allow_warnings
        self._endpoint = _endpoint
        self._next_link = None
        self._prev_link = None
        self._next_page = False
        self._prev_page = False
        

    def _ret(self) -> 'Api':
        """
        Функция для создания нового класса, с новыми данными, после каждого добавленного условия. 
        (Вынесено отдельно, чтобы не загружать каждую функцию одинаковым кодом)
        """
        return Api(allow_warnings=self.allow_warnings, _args=self._args, _endpoint=self._endpoint)
    
    def title(self, title: str) -> 'Api':
        """
        Указывается название мультимедиа для поиска.
        :title: Название.
        """
        self._args['title'] = title
        return self._ret()
    
    def title_orig(self, title_orig: str) -> 'Api':
        """
        Указывается оригинальное название мультимедиа для поиска. Обычно таким считается название на английском (Даже если оригинально названо на русском).
        :title_orig: Оригинально название.
        """
        self._args['title_orig'] = title_orig
        return self._ret()
    
    def strict(self, strict: bool = True) -> 'Api':
        """
        Параметр strict отвечает за более точное совпадение запрошенного названия/оригинального названия с найденными.
        :strict: Более точное совпадение. По умолчанию True.
        """
        self._args['strict'] = strict
        return self._ret()
    
    def full_match(self, full_match: bool = True) -> 'Api':
        """
        Требует полного совпадения запрошенного названия с найденным (зависит от регистра).
        :full_match: Полное совпадение. По умолчанию True.
        """
        self._args['full_match'] = full_match
        return self._ret()
    
    def shikimori_id(self, shikimori_id: str) -> 'Api':
        """
        Поиск по id шикимори.
        :shikimori_id: id шикимори.
        
        Как узнать id:
        1. Перейдите на страницу аниме на шикимори
        2. В адресной строке сайта будет примерно следующее: 
            https://shikimori.one/animes/52991-sousou-no-frieren
        3. Строка (обычно цифры, но могут встречаться и буквы) после последнего "/" и перед первым "-" и будет являться shikimori_id
            В данном случае: 52991 
        """
        self._args['shikimori_id'] = shikimori_id
        return self._ret()
    
    def kinopoisk_id(self, kinopoisk_id: str) -> 'Api':
        """
        Поиск по id кинопоиска.
        :kinopoisk_id: id кинопоиска.
        
        Как узнать id:
        1. Перейдите на страницу фильма/сериала на кинопоиске
        2. В адресной строке сайта будет примерно следующее: 
            https://www.kinopoisk.ru/film/258687/
        3. Строка (обычно цифры, но могут встречаться и буквы) после пред-последнего "/" и перед последним "/" и будет являться kinopoisk_id
            В данном случае: 258687
        """
        self._args['kinopoisk_id'] = kinopoisk_id
        return self._ret()
    
    def imdb_id(self, imdb_id: str) -> 'Api':
        """
        Поиск по id imdb.
        :imdb_id: id imdb.
        
        Как узнать id:
        1. Перейдите на страницу фильма/сериала на imdb
        2. В адресной строке сайта будет примерно следующее: 
            https://www.imdb.com/title/tt0816692/
        3. Строка (обычно содержит в начале "tt") после пред-последнего "/" и перед последним "/" и будет являться imdb_id
            В данном случае: tt0816692
        """
        self._args['imdb_id'] = imdb_id
        return self._ret()
    
    def id(self, id: str) -> 'Api':
        """
        Поиск по внутреннему id кодика.
        :id: Внутренний id кодика.

        Как узнать id: При выполнении запроса, в каждом элементе будет находится поле id.
        """
        self._args['id'] = id
        return self._ret()

    def mdl_id(self, mdl_id: str) -> 'Api':
        """
        Поиск по id MyDramaList.
        :mdl_id: id MyDramaList.
        
        Как узнать id:
        1. Перейдите на страницу фильма/сериала на mydramalist.com
        2. В адресной строке сайта будет примерно следующее: 
            https://mydramalist.com/2570-space-battleship-yamato
        3. Строка (обычно цифры, но могут встречаться и буквы) после последнего "/" и перед первым "-" и будет являться mdl_id
            В данном случае: 2570
        """
        self._args['mdl_id'] = mdl_id
        return self._ret()
    
    def worldart_animation_id(self, worldart_animation_id: str) -> 'Api':
        """
        Поиск по id Worldart Animation.
        :worldart_animation_id: id Worldart Animation.
        
        Как узнать id:
        1. Перейдите на страницу фильма/сериала на www.world-art.ru/anime
        2. В адресной строке сайта будет примерно следующее: 
            http://www.world-art.ru/animation/animation.php?id=11466
        3. Строка (обычно цифры, но могут встречаться и буквы) после "id=" является искомым id
            В данном случае: 11466
        """
        self._args['worldart_animation_id'] = worldart_animation_id
        return self._ret()
    
    def worldart_cinema_id(self, worldart_cinema_id: str) -> 'Api':
        """
        Поиск по id Worldart Cinema.
        :worldart_cinema_id: id Worldart Cinema.
        
        Как узнать id:
        1. Перейдите на страницу фильма/сериала на www.world-art.ru/cinema
        2. В адресной строке сайта будет примерно следующее: 
            http://www.world-art.ru/cinema/cinema.php?id=36765
        3. Строка (обычно цифры, но могут встречаться и буквы) после "id=" является искомым id
            В данном случае: 36765
        """
        self._args['worldart_cinema_id'] = worldart_cinema_id
        return self._ret()
    
    def worldart_link(self, worldart_link: str) -> 'Api':
        """
        Ссылка на страницу world-art.
        :worldart_link: Ссылка на страницу world-art.

        Пример: http://www.world-art.ru/animation/animation.php?id=11466
        """
        self._args['worldart_link'] = worldart_link
        return self._ret()
    
    def limit(self, limit: int = 50) -> 'Api':
        """
        Установка предела на количество элементов в ответе.
        Допустимые пределы: 1-100 (Если меньше 1, то сервер вернет ошибку 500, если больше 100, то сервер сам ограничит количество до 100)
        :limit: Предел на количество элементов. По умолчанию 50.
        """
        if limit < 1:
            raise ValueError("Параметр limit должен быть указан в пределах 1-100! В случае если limit будет меньше одного, сервер гарантированно вернет ошибку 500.")
        elif limit > 100 and self.allow_warnings:
            print('Предупреждение! Параметр limit установлен больше 100, сервер не вернет больше 100 элементов.')
        self._args['limit'] = limit
        return self._ret()
    
    def types(self, types: list | str) -> 'Api':
        """
        Типы мультимедиа. Можно передавать либо один вариант, либо несколько.
        :types: Типы мультимедиа. Строка или список строк.

        Вы можете воспользоваться вспомогательным классом Types (Search.Types или List.Types или Api.Types).
        Доступные фильтры:

        * foreign-movie - Иностранный фильм
        * soviet-cartoon - Советский мультфильм
        * foreign-cartoon - Иностранный мультфильм
        * russian-cartoon - Русский мультфильм
        * anime - Аниме фильм (либо одна серия)
        * russian-movie - Русский фильм
        * cartoon-serial - мульт-сериал
        * documentary-serial - документальный сериал
        * russian-serial - Русский сериал
        * foreign-serial - Иностранный сериал
        * anime-serial - Аниме сериал (все, что больше 1 серии)
        * multi-part-film - Фильм из нескольких частей
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)
        """
        if type(types) == list:
            if self.allow_warnings and any([x in self.Types.get_list() for x in types]):
                print('Предупреждение! Один из параметров types не содержится в списках доступных!')
            self._args['types'] = ','.join(types)
        elif type(types) == str:
            if self.allow_warnings and types not in self.Types.get_list():
                print(f'Предупреждение! Один из параметров types ("{types}") не содержится в списках доступных!')
            self._args['types'] = types
        else:
            raise ValueError(f"Ожидался тип данных 'list' или 'str'. Получено: '{type(types)}'")
        return self._ret()

    def year(self, year: int) -> 'Api':
        """
        Год выпуска мультимедиа.
        :year: Год выпуска мультимедиа. Целое число.
        """
        self._args['year'] = year
        return self._ret()
    
    def camrip(self, camrip: bool = True) -> 'Api':
        """
        Снято ли с камеры.
        :camrip: Снято ли с камеры. По умолчанию True
        """
        self._args['camrip'] = camrip
        return self._ret()
    
    def lgbt(self, lgbt: bool = True) -> 'Api':
        """
        Имеется ли lgbt контент.
        :camrip: Имеется ли lgbt контент. По умолчанию True
        """
        self._args['lgbt'] = lgbt
        return self._ret()
    
    def translation_id(self, translation_id: int) -> 'Api':
        """
        Id перевода. Посмотреть все доступные переводы можно используя эндпоинт translations
        :translation_id: Id перевода.
        """
        self._args['translation_id'] = translation_id
        return self._ret()
    
    def translation_type(self, translation_type: str) -> 'Api':
        """
        Тип перевода. Озвучка или субтитры. subtitles/voice
        :translation_type: Тип перевода. subtitles/voice
        """
        if translation_type not in ['subtitles', 'voice']:
            raise ValueError(f'Параметр translation_type должен быть "subtitles" или "voice", получено: "{translation_type}"')
        self._args['translation_type'] = translation_type
        return self._ret()
    
    def anime_kind(self, anime_kind: str) -> 'Api':
        """
        Тип аниме (указывая этот параметр, в результате будет только аниме).
        :anime_kind: Тип аниме.

        Вы можете воспользоваться вспомогательным классом AnimeKind (KodikSearch.AnimeKind или KodikList.AnimeKind или Api.AnimeKind).
        Доступные фильтры:\n
        * tv - ТВ Сериал
        * tv13 - Длительность серии ~13 минут
        * tv24 - Длительность серии ~24 минут
        * tv48 - Длительность серии ~48 минут
        * movie - Фильм
        * special - Спец. выпуск
        * ova
        * ona
        * music - Музыкальный клип / Проморолик
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)\n
        """
        if self.allow_warnings and anime_kind not in self.AnimeKind.get_list():
            print(f'Предупреждение! Параметр anime_kind "{anime_kind}" не содержится в списках доступных.')
        self._args['anime_kind'] = anime_kind
        return self._ret()
    
    def anime_status(self, anime_status: str) -> 'Api':
        """
        Статус выхода аниме. Вышло/Онгоинг/Анонс - released/ongoing/anons
        :anime_status: Статус выхода аниме.
        """
        if anime_status not in ['released', 'ongoing', 'anons']:
            raise ValueError(f'Параметр anime_status должен быть одним из "released", "ongoing", "anons". Получено: {anime_status}')
        self._args['anime_status'] = anime_status
        return self._ret()
    
    def mydramalist_tags(self, mydramalist_tags: list|str) -> 'Api':
        """
        Теги на MyDramaList.
        :mydramalist_tags: Теги на MyDramaList.
        """
        if type(mydramalist_tags) == list:
            self._args['mydramalist_tags'] = ','.join(mydramalist_tags)
        else:
            self._args['mydramalist_tags'] = mydramalist_tags
        return self._ret()
    
    def rating_mpaa(self, rating_mpaa: str) -> 'Api':
        """
        Возрастной рейтинг по mpaa.
        :rating_mpaa: Возрастной рейтинг по mpaa.

        Вы можете воспользоваться вспомогательным классом MPAArating (KodikSearch.MPAArating или KodikList.MPAArating или Api.MPAArating).
        Доступные фильтры:\n
        * g - Нет возрастных ограничений.
        * pg - Рекомендуется присутствие родителей.
        * pg-13 - Детям до 13 лет просмотр нежелателен.
        * r - Лицам до 17 лет обязательно присутствие взрослого.
        * rx - Хeнтай / Пoрнография    
        \nЧисто технически, существует рейтинг r+, но почему-то у меня не получилось отфильтровать по нему поиск (выдавался рейтинг r).
        \n(Для передачи параметра напрямую, строкой, укажите его на английском, как указано выше)
        """
        if self.allow_warnings and rating_mpaa not in self.MPAArating.get_list():
            print(f'Предупреждение! Параметр rating_mpaa "{rating_mpaa}" не содержится в доступных.')
        self._args['rating_mpaa'] = rating_mpaa
        return self._ret()
    
    def minimal_age(self, minimal_age: int) -> 'Api':
        """
        Минимальный возраст зрителя. Через фильтр проходят все элементы, у которых указанный возраст совпадает с переданным параметром или больше его.
        :minimal_age: Минимальный возраст зрителя.
        """
        self._args['minimal_age'] = minimal_age
        return self._ret()
    
    def kinopoisk_rating(self, kinopoisk_rating: float) -> 'Api':
        """
        Рейтинг на кинопоиске. Ищет только элементы, у которых оценка точно совпадает с указанным.
        :kinopoisk_rating: Рейтинг на кинопоиске.
        """
        self._args['kinopoisk_rating'] = kinopoisk_rating
        return self._ret()
    
    def imdb_rating(self, imdb_rating: float) -> 'Api':
        """
        Рейтинг на imdb. Ищет только элементы, у которых оценка точно совпадает с указанным.
        :imdb_rating: Рейтинг на imdb.
        """
        self._args['imdb_rating'] = imdb_rating
        return self._ret()
    
    def shikimori_rating(self, shikimori_rating: float) -> 'Api':
        """
        Рейтинг на шикимори. Ищет только элементы, у которых оценка точно совпадает с указанным.
        :shikimori_rating: Рейтинг на шикимори.
        """
        self._args['shikimori_rating'] = shikimori_rating
        return self._ret()
    
    def anime_studios(self, anime_studios: list|str) -> 'Api':
        """
        Студия/студии анимации занимающиеся производством. Зависит от регистра.
        :anime_studios: Студия/студии анимации занимающиеся производством.
        """
        if type(anime_studios) == list:
            self._args['anime_studios'] = ','.join(anime_studios)
        else:
            self._args['anime_studios'] = anime_studios
        return self._ret()
    
    def genres(self, genres: str | list) -> 'Api':
        """
        Жанры мультимедиа. Принимает строку или список строк.
        :genres: Жанры мультимедиа.

        Вы можете воспользоваться вспомогательным классом Genres (KodikSearch.Genres или KodikList.Genres или Api.Genres).
        Доступные фильтры:\n    
        * anime - аниме
        * drama - драма
        * comedy - комедия
        * cartoon - мультфильм
        * fantasy - фэнтези
        * action - боевик
        * adventures - приключения
        * fantastika - фантастика
        \n(Для передачи параметра напрямую, строкой, укажите его на русском, как указано выше)
        """
        if type(genres) == list:
            if self.allow_warnings and any([x not in self.Genres.get_list() for x in genres]):
                print('Предупреждение! Один из указанных жанров отсутствует в списках доступных.')
            self._args['genres'] = ','.join(genres)
        else:
            if self.allow_warnings and genres not in self.Genres.get_list():
                print(f'Предупреждение! Указанный жанр "{genres}" отсутствует в списках доступных.')
            self._args['genres'] = genres
        return self._ret()
    
    def anime_genres(self, anime_genres: str | list) -> 'Api':
        """
        Аниме жанры. Принимает строку или список строк.
        :anime_genres: Аниме жанры.

        Вы можете воспользоваться вспомогательным классом AnimeGenres (KodikSearch.AnimeGenres или KodikList.AnimeGenres или Api.AnimeGenres).
        Доступные фильтры:\n
        * military - Военное
        * drama - Драма
        * history - Исторический
        * action - Экшен
        * adventures - Приключения
        * senen - Сёнен
        * fantasy - Фэнтези
        * comedy - Комедия
        * martial_arts - Боевые искусства
        * romance - Романтика
        * psychological - Психологическое
        * thriller - Триллер
        * everyday_life - Повседневность
        * supernatural - Сверхъестественное
        * sport - Спорт
        * school - Школа
        * music - Музыка
        * fantastika - Фантастика
        * samurai - Самураи
        \n(Для передачи параметра напрямую, строкой, укажите его на русском, как указано выше)
        """
        if type(anime_genres) == list:
            if self.allow_warnings and any([x not in self.AnimeGenres.get_list() for x in anime_genres]):
                print('Предупреждение! Один из указанных аниме жанров отсутствует в списках доступных.')
            self._args['anime_genres'] = ','.join(anime_genres)
        else:
            if self.allow_warnings and anime_genres not in self.AnimeGenres.get_list():
                print(f'Предупреждение! Указанный аниме жанр "{anime_genres}" отсутствует в списках доступных.')
            self._args['anime_genres'] = anime_genres
        return self._ret()
    
    def duration(self, duration: int) -> 'Api':
        """
        Продолжительность одной серии или фильма в минутах. Возвращаются только элементы с полностью совпадающим значением.
        :duration: Продолжительность в минутах.
        """
        self._args['duration'] = duration
        return self._ret()
    
    def player_link(self, player_link: str) -> 'Api':
        """
        Ссылка на embed страницу кодика. Пример: kodik.info/serial/48654/3520f3c13f024b8fe04bc6143ccdcb7d/720p
        :player_link: Ссылка на embed кодика.
        """
        self._args['player_link'] = player_link
        return self._ret()
    
    def has_field(self, has_field: str) -> 'Api':
        """
        Элемент содержит указанный параметр.
        """
        self._args['has_field'] = has_field
        return self._ret()
    
    def has_fields(self, has_fields: list) -> 'Api':
        """
        Элемент содержит указанные параметры.
        :has_fields: Список строк
        """
        self._args['has_fields'] = ','.join(has_fields)
        return self._ret()
    
    def has_field_and(self, has_field_and: list) -> 'Api':
        """
        Элемент содержит указанные параметры.
        :has_field_and: Список строк
        """
        self._args['has_field_and'] = ','.join(has_field_and)
        return self._ret()
    
    def prioritize_translations(self, prioritize_translations: str | list) -> 'Api':
        """
        Повысить приоритет указанных id переводов в результатах выдачи. Принимает строку или список строк.
        :prioritize_translations: Строка или список строк с id переводов.

        Посмотреть список переводов можно по эндпоинту translations. 
        По умолчанию приоритет выдается профессиональным и многоголосым озвучкам.
        Для отключения стандартного приоритета укажите 0.
        """
        if type(prioritize_translations) == list:
            self._args['prioritize_translations'] = ','.join(prioritize_translations)
        else:
            self._args['prioritize_translations'] = prioritize_translations
        return self._ret()
    
    def unprioritize_translations(self, unprioritize_translations: str | list) -> 'Api':
        """
        Понизить приоритет указанных id переводов в результатах выдачи. Принимает строку или список строк.
        :unprioritize_translations: Строка или список строк с id переводов.

        Посмотреть список переводов можно по эндпоинту translations. 
        По умолчанию установлен низкий приоритет озвучкам на Английском и Украинском языках, а также субтитров.
        Для отключения стандартного приоритета укажите 0.
        """
        if type(unprioritize_translations) == list:
            self._args['unprioritize_translations'] = ','.join(unprioritize_translations)
        else:
            self._args['unprioritize_translations'] = unprioritize_translations
        return self._ret()
    
    def block_translations(self, block_translations: str | list) -> 'Api':
        """
        Убрать переводы из результатов. Принимает строку или список строк.
        :block_translations: Строка или список строк с id переводов. 
        """
        if type(block_translations) == list:
            self._args['block_translations'] = ','.join(block_translations)
        else:
            self._args['block_translations'] = block_translations
        return self._ret()
    
    def prioritize_translation_type(self, prioritize_translation_type: str) -> 'Api':
        """
        Увеличить приоритет типа перевода subtitles/voice.
        :prioritize_translation_type: subtitles/voice
        """
        if prioritize_translation_type not in ['subtitles', 'voice']:
            raise ValueError(f'Параметр prioritize_translation_type должен быть "subtitles" или "voice". Получено: "{prioritize_translation_type}"')
        self._args['prioritize_translation_type'] = prioritize_translation_type
        return self._ret()
    
    def season(self, season: int) -> 'Api':
        """
        Сезон выхода аниме.
        :season: Сезон выхода аниме.
        """
        self._args['season'] = season
        return self._ret()
    
    def episode(self, episode: int) -> 'Api':
        """
        В результате будут только элементы у которых есть указанный эпизод.
        :episode: Эпизод для поиска.
        """
        self._args['episode'] = episode
        return self._ret()
    
    def not_blocked_in(self, not_blocked_in: str | list) -> 'Api':
        """
        В результате будут элементы не заблокированные в указанной стране/странах.
        :not_blocked_in: Строка или список строк с названиям стран на русском языке (зависит от регистра).
        """
        if type(not_blocked_in) == list:
            self._args['not_blocked_in'] = ','.join(not_blocked_in)
        else:
            self._args['not_blocked_in'] = not_blocked_in
        return self._ret()
    
    def not_blocked_for_me(self, not_blocked_for_me: bool = True) -> 'Api':
        """
        В результате будут элементы не заблокированные в стране, откуда отправлен запрос (предположительно проверка по ip)
        :not_blocked_for_me: Не заблокировано в стране запроса. По умолчанию True.
        """
        self._args['not_blocked_for_me'] = not_blocked_for_me
        return self._ret()
    
    def countries(self, countries: str | list) -> 'Api':
        """
        Страны в которых снят фильм/сериал.
        :countries: Строка или список строк с названиям стран на русском языке (зависит от регистра).
        """
        if type(countries) == list:
            self._args['countries'] = ','.join(countries)
        else:
            self._args['countries'] = countries
        return self._ret()
    
    def actors(self, actors: str | list) -> 'Api':
        """
        Вернет элементы где в актерах указан(ы) переданные имена.
        :actors: Строка или список строк с именами (зависит от регистра).
        """
        if type(actors) == list:
            self._args['actors'] = ','.join(actors)
        else:
            self._args['actors'] = actors
        return self._ret()
    
    def directors(self, directors: str | list) -> 'Api':
        """
        Вернет элементы где в режиссерах указан(ы) переданные имена.
        :directors: Строка или список строк с именами (зависит от регистра).
        """
        if type(directors) == list:
            self._args['directors'] = ','.join(directors)
        else:
            self._args['directors'] = directors
        return self._ret()
    
    def producers(self, producers: str | list) -> 'Api':
        """
        Вернет элементы где в продюсерах указан(ы) переданные имена.
        :producers: Строка или список строк с именами (зависит от регистра).
        """
        if type(producers) == list:
            self._args['producers'] = ','.join(producers)
        else:
            self._args['producers'] = producers
        return self._ret()
    
    def writers(self, writers: str | list) -> 'Api':
        """
        Вернет элементы где в писателях/сценаристах указан(ы) переданные имена.
        :writers: Строка или список строк с именами (зависит от регистра).
        """
        if type(writers) == list:
            self._args['writers'] = ','.join(writers)
        else:
            self._args['writers'] = writers
        return self._ret()
    
    def composers(self, composers: str | list) -> 'Api':
        """
        Вернет элементы где в композиторах указан(ы) переданные имена.
        :composers: Строка или список строк с именами (зависит от регистра).
        """
        if type(composers) == list:
            self._args['composers'] = ','.join(composers)
        else:
            self._args['composers'] = composers
        return self._ret()

    def editors(self, editors: str | list) -> 'Api':
        """
        Вернет элементы где в монтажерах указан(ы) переданные имена.
        :editors: Строка или список строк с именами (зависит от регистра).
        """
        if type(editors) == list:
            self._args['editors'] = ','.join(editors)
        else:
            self._args['editors'] = editors
        return self._ret()
    
    def designers(self, designers: str | list) -> 'Api':
        """
        Вернет элементы где в дизайнерах/художниках указан(ы) переданные имена.
        :designers: Строка или список строк с именами (зависит от регистра).
        """
        if type(designers) == list:
            self._args['designers'] = ','.join(designers)
        else:
            self._args['designers'] = designers
        return self._ret()
    
    def operators(self, operators: str | list) -> 'Api':
        """
        Вернет элементы где в операторах указан(ы) переданные имена.
        :operators: Строка или список строк с именами (зависит от регистра).
        """
        if type(operators) == list:
            self._args['operators'] = ','.join(operators)
        else:
            self._args['operators'] = operators
        return self._ret()
    
    def licensed_by(self, licensed_by: str | list) -> 'Api':
        """
        Вернет элементы где владельцем лицензии (на перевод/озвучку ?) является указанная компания.
        :licensed_by: Строка или список строк с названиями компаний - владельцев лицензии. (Зависит от регистра). 
        """
        if type(licensed_by) == list:
            self._args['licensed_by'] = ','.join(licensed_by)
        else:
            self._args['licensed_by'] = licensed_by
        return self._ret()
    
    def with_material_data(self, with_material_data: bool = True) -> 'Api':
        """
        Добавляет в результат дополнительные данные о мультимедиа.
        :with_material_data: Добавить доп. данные. По умолчанию True.
        """
        self._args['with_material_data'] = with_material_data
        return self._ret()
    
    def with_seasons(self, with_seasons: bool = True) -> 'Api':
        """
        Добавляет в результат ссылки на сезоны. (ссылки на embed на kodik.info)
        :with_seasons: Добавить список сезонов. По умолчанию True.
        """
        self._args['with_seasons'] = with_seasons
        return self._ret()
    
    def with_episodes(self, with_episodes: bool = True) -> 'Api':
        """
        Добавляет в результат ссылки на конкретные эпизоды, автоматически включается параметр with_seasons. (ссылки на embed на kodik.info)
        :with_episodes: Добавить список эпизодов. По умолчанию True.
        """
        self._args['with_episodes'] = with_episodes
        return self._ret()
    
    def with_episodes_data(self, with_episodes_data: bool = True) -> 'Api':
        """
        Добавляет в результат данные для конкретных эпизодов (если есть) в виде названия, описания, автоматически включается параметр with_episodes.
        :with_episodes_data: Добавить данные для эпизодов, если есть. По умолчанию True.
        """
        self._args['with_episodes_data'] = with_episodes_data
        return self._ret()
    
    def with_page_links(self, with_page_links: bool = True) -> 'Api':
        """
        Заменяет в результате ссылку на embed (kodik.info) на сайт с плеером (kodik.online)
        :with_page_links: Заменить ссылку на сайт с плеером.
        """
        self._args['with_page_links'] = with_page_links
        return self._ret()
    
    # ================================================================
    def api_request(self, parameters: dict | None = None, link: str | None = None) -> dict:
        """
        Запрос к api. Требуется передать только один из параметров parameters или link

        :parameters: Параметры и фильтры для запроса.
        :link: Полная ссылка (для следующей и предыдущей страниц)
        """
        if link is None and self._endpoint not in ['search', 'list', 'translations']:
            raise errors.PostArgumentsError(f'Неизвестный эндпоинт. Ожидался один из "search", "list", "translations". Получен: "{self._endpoint}"')
        if self.token is None:
            raise errors.TokenError('Токен kodik не указан')
        if parameters is None and link is None:
            raise ValueError('Хотя-бы один из параметров "parameters" или "link" должен быть указан для использования api_request')
        
        if parameters:
            payload = {"token": self.token}
            for item, val in parameters.items():
                payload[item] = val

            url = f"https://kodikapi.com/{self._endpoint}"
            data = requests.post(url, data=payload)
        else:
            url = link
            data = requests.post(url)
        
        if data.status_code != 200:
            raise errors.ServiceError(f'Произошла ошибка при запросе к kodik api. Ожидался код "200", получен: "{data.status_code}"')
        data = data.json()
        
        if 'error' in data.keys() and data['error'] == 'Отсутствует или неверный токен':
            raise errors.TokenError('Отсутствует или неверный токен')
        elif 'error' in data.keys():
            raise errors.ServiceError(data['error'])
        if data['total'] == 0:
            raise errors.NoResults('Сервер вернул ответ с пустым списком результатов.')
        if 'next_page' in data.keys():
            self._next_link = data['next_page']
        else:
            self._next_link = None
        if 'prev_page' in data.keys():
            self._prev_link = data['prev_page']
        else:
            self._prev_link = None
        return data
    
    async def api_request_async(self, parameters: dict | None = None, link: str | None = None) -> dict:
        """
        Асинхронный запрос к api. Требуется передать только один из параметров parameters или link

        :parameters: Параметры и фильтры для запроса.
        :link: Полная ссылка (для следующей и предыдущей страниц)
        """
        if not async_available:
            raise RuntimeError('При попытке импорта внутренних инструментов библиотеки anime-parsers-ru произошла ошибка и AsyncSession не доступен.')

        if link is None and self._endpoint not in ['search', 'list', 'translations']:
            raise errors.PostArgumentsError(f'Неизвестный эндпоинт. Ожидался один из "search", "list", "translations". Получен: "{self._endpoint}"')
        if self.token is None:
            raise errors.TokenError('Токен kodik не указан')
        if parameters is None and link is None:
            raise ValueError('Хотя-бы один из параметров "parameters" или "link" должен быть указан для использования api_request')
        
        if parameters:
            payload = {"token": self.token}
            for item, val in parameters.items():
                payload[item] = val
            url = f"https://kodikapi.com/{self._endpoint}"
            data = await AsyncSession().post(url, data=payload)
        else:
            url = link
            data = await AsyncSession().post(url)
        
        if data.status_code != 200:
            raise errors.ServiceError(f'Произошла ошибка при запросе к kodik api. Ожидался код "200", получен: "{data.status_code}"')
        data = data.json()
        
        if 'error' in data.keys() and data['error'] == 'Отсутствует или неверный токен':
            raise errors.TokenError('Отсутствует или неверный токен')
        elif 'error' in data.keys():
            raise errors.ServiceError(data['error'])
        if data['total'] == 0:
            raise errors.NoResults('Сервер вернул ответ с пустым списком результатов.')
        if 'next_page' in data.keys():
            self._next_link = data['next_page']
        else:
            self._next_link = None
        if 'prev_page' in data.keys():
            self._prev_link = data['prev_page']
        else:
            self._prev_link = None
        return data

    def execute(self, return_json: bool = False) -> Response:
        """
        Выполнить запрос.
        :return_json: Возвращать чистый ответ сервера, вместо Response. По умолчанию False.
        """
        if self._endpoint is None:
            raise TypeError('Don\'t use _ApiRequest directly! Use Search and List classes instead.')
        data = self.api_request(parameters=self._args)
        if return_json:
            return data
        else:
            return Response(data)

    async def execute_async(self, return_json: bool = False) -> Response:
        """
        Выполнить запрос асинхронно.
        :return_json: Возвращать чистый ответ сервера, вместо Response. По умолчанию False.
        """
        if self._endpoint is None:
            raise TypeError('Don\'t use _ApiRequest directly! Use Search and List classes instead.')
        data = await self.api_request_async(parameters=self._args)
        if return_json:
            return data
        else:
            return Response(data)
    
    def next_page(self, link: str | None = None) -> Response:
        """
        Выполнить запрос для получения следующей страницы списка. (Не работает для поиска)

        Для использования требуется выполнение обычного запроса типа List который в своем ответе вернет ссылку на следующую страницу.
        Или передать параметр link.

        :link: Полная ссылка на следующую страницу. Если None - использует внутреннюю переменную на следующую страницу. По умолчанию None.
        """
        if link:
            data = self.api_request(link=link)
            return Response(data)
        if self._next_link:
            data = self.api_request(link=self._next_link)
            return Response(data)
        raise ValueError('Для использования этой функции параметр next_page должен быть в результатах предыдущего запроса.')
    
    def prev_page(self, link: str | None = None) -> Response:
        """
        Выполнить запрос для получения предыдущей страницы списка. (Не работает для поиска)

        Для использования требуется выполнение обычного запроса типа List который в своем ответе вернет ссылку на предыдущую страницу.
        Или передать параметр link.

        :link: Полная ссылка на предыдущую страницу. Если None - использует внутреннюю переменную на предыдущую страницу. По умолчанию None.
        """
        if link:
            data = self.api_request(link=link)
            return Response(data)
        if self._prev_link:
            data = self.api_request(link=self._prev_link)
            return Response(data)
        raise ValueError('Для использования этой функции параметр prev_page должен быть в результатах предыдущего запроса.')
    
    async def next_page_async(self, link: str | None = None) -> Response:
        """
        Выполнить асинхронный запрос для получения следующей страницы списка. (Не работает для поиска)

        Для использования требуется выполнение обычного запроса типа List который в своем ответе вернет ссылку на следующую страницу.
        Или передать параметр link.

        :link: Полная ссылка на следующую страницу. Если None - использует внутреннюю переменную на следующую страницу. По умолчанию None.
        """
        if link:
            data = await self.api_request_async(link=link)
            return Response(data)
        if self._next_link:
            data = await self.api_request_async(link=self._next_link)
            return Response(data)
        raise ValueError('Для использования этой функции параметр next_page должен быть в результатах предыдущего запроса.')
    
    async def prev_page_async(self, link: str | None = None) -> Response:
        """
        Выполнить асинхронный запрос для получения предыдущей страницы списка. (Не работает для поиска)

        Для использования требуется выполнение обычного запроса типа List который в своем ответе вернет ссылку на предыдущую страницу.
        Или передать параметр link.

        :link: Полная ссылка на предыдущую страницу. Если None - использует внутреннюю переменную на предыдущую страницу. По умолчанию None.
        """
        if link:
            data = await self.api_request_async(link=link)
            return Response(data)
        if self._prev_link:
            data = await self.api_request_async(link=self._prev_link)
            return Response(data)
        raise ValueError('Для использования этой функции параметр prev_page должен быть в результатах предыдущего запроса.')
    
    def get_translations(self) -> dict:
        """
        Выполняет запрос к апи по эндпоинту translations.
        Возвращает словарь.
        """
        self._endpoint = 'translations' # Да, костыль. Но плодить еще один класс чтобы сделать 1 запрос, тоже странно.
        data = self.api_request(parameters={})
        self._endpoint = None
        return data


class KodikSearch(Api):
    """
    Абстракция для запроса к апи плеера кодик.

    Документацию по доступным параметрам вы можете найти в репозитории: https://github.com/YaNesyTortiK/AnimeParsers/blob/main/KODIK_API.md
    
    Пример использования:

    query = Search().title("Кулинарные скитания").limit(5).anime_status('released')
    data = query.execute()

    """
    def __init__(self, token: str | None = None, allow_warnings: bool = True, _args: dict = {}):
        super().__init__(token, allow_warnings, _args, _endpoint='search')

class KodikList(Api):
    def __init__(self, token: str | None = None, allow_warnings: bool = True, _args: dict = {}):
        super().__init__(token, allow_warnings, _args, _endpoint='list')
    
    def title(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр title не доступен.')
    
    def title_orig(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр title_orig не доступен.')
    
    def strict(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр strict не доступен.')

    def full_match(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр full_match не доступен.') 

    def shikimori_id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр shikimori_id не доступен.') 

    def kinopoisk_id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр kinopoisk_id не доступен.') 

    def imdb_id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр imdb_id не доступен.') 

    def id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр id не доступен.') 

    def mdl_id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр mdl_id не доступен.') 

    def worldart_cinema_id(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр worldart_cinema_id не доступен.') 

    def worldart_link(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр worldart_link не доступен.') 
    
    def player_link(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр player_link не доступен.') 
    
    def prioritize_translations(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр prioritize_translations не доступен.') 
    
    def unprioritize_translations(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр unprioritize_translations не доступен.') 
    
    def season(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр season не доступен.') 
    
    def episode(self, *args):
        """
        Данный параметр недоступен для запроса list.
        """
        raise errors.PostArgumentsError('Для запроса list параметр episode не доступен.') 
    
